from __future__ import annotations

from typing import TypeVar, Generic, Literal, get_type_hints
import enum

from cohdl import (
    Null,
    Full,
    Signal,
    Variable,
    BitVector,
    Unsigned,
    Signed,
    Bit,
    Temporary,
    expr_fn,
    always,
)
from cohdl import std
from cohdl._core import AssignMode, Array, true
from cohdl._core._intrinsic import _intrinsic
from cohdl.utility import TextBlock

from .._context import SequentialContext, concurrent
from ..enum import Enum as StdEnum
from .._core_utility import to_bits, NoresetSignal, zeros, ones, concat, base_type
from ..utility import is_pow_two, int_log_2


class Access(enum.Enum):
    na = enum.auto()
    r = enum.auto()
    w = enum.auto()
    w1 = enum.auto()
    rw = enum.auto()
    rw1 = enum.auto()


class HwAccess(enum.Enum):
    na = enum.auto()
    r = enum.auto()
    w = enum.auto()
    rw = enum.auto()


def _expand_lists(inp):
    if isinstance(inp, list):
        result = []

        for elem in inp:
            if isinstance(elem, list):
                result.extend(_expand_lists(elem))
            else:
                result.append(elem)

        return result

    return [inp]


class _None: ...


class _Infinite:
    # The word_count argument is optional for the root AddrMap.
    # When it is not specified an instance of _Infinite is used
    # to pass all bound checks.

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __gt__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __lt__(self, other):
        return False


class GenericArg:
    offset: int = None
    end: int | None = None
    array_step: int | None = None

    def __init__(self, arg):
        self.array_type = None

        if isinstance(arg, tuple):
            array_type, arg = arg
            assert issubclass(array_type, RegisterTools.RegisterObject)
            self.array_type = array_type

        if isinstance(arg, int):
            self.offset = arg
            self.end = None
            self.array_step = None
        elif isinstance(arg, slice):
            assert isinstance(arg.start, int)
            self.offset = arg.start
            self.end = arg.stop
            self.array_step = arg.step

    def __hash__(self) -> int:
        return hash((self.array_type, self.offset, self.end, self.array_step))

    def __eq__(self, other: GenericArg):
        return (
            self.array_type is other.array_type
            and self.offset == other.offset
            and self.end == other.end
            and self.array_step == other.array_step
        )

    def __str__(self):
        return f"{self.offset}"


def _config_wrapper(self, *args, **kwargs):
    assert not hasattr(self, "_cohdlstd_objconfigured")
    result = self._cohdlstd_wrappedconfig(*args, **kwargs)
    setattr(self, "_cohdlstd_objconfigured", True)
    return result


class _RegisterToolsArg:
    word_width: int
    unit_width: int
    word_stride: int

    Field: type[std.bitfield.Field]

    def __init__(self, arg):
        if not isinstance(arg, tuple):
            word_width = arg
            unit_width = 8
        else:
            word_width, unit_width = arg

        assert isinstance(word_width, int)
        assert isinstance(unit_width, int)
        assert (
            word_width % unit_width == 0
        ), "word with must be a multiple of the unit width"

        self.word_width = word_width
        self.unit_width = unit_width
        self.word_stride = word_width // unit_width

    def __hash__(self) -> int:
        return hash((self.word_width, self.unit_width))

    def __eq__(self, other: _RegisterToolsArg) -> bool:
        return (
            self.word_width == other.word_width and self.unit_width == other.unit_width
        )


class RegisterTools(std.Template[_RegisterToolsArg]):
    _word_width_: _RegisterToolsArg.word_width
    _addr_unit_width_: _RegisterToolsArg.unit_width
    _word_stride_: _RegisterToolsArg.word_stride

    @classmethod
    def _as_word_addr_(cls, unit_addr: Unsigned):
        assert is_pow_two(cls._word_stride_)
        drop_bits = int_log_2(cls._word_stride_)
        return unit_addr.msb(rest=drop_bits).copy().unsigned

    @classmethod
    def _as_unit_addr_(cls, word_addr: Unsigned):
        assert is_pow_two(cls._word_stride_)
        add_bits = int_log_2(cls._word_stride_)
        return (word_addr @ zeros(add_bits)).unsigned

    @classmethod
    def _template_specialize_(cls):
        for name in dir(cls):
            subtype = getattr(cls, name)

            if not isinstance(subtype, type):
                continue

            if issubclass(subtype, AddrMap):
                continue

            if issubclass(subtype, (RegFile, AddrRange)):
                setattr(
                    cls,
                    name,
                    type(
                        subtype.__name__,
                        (subtype,),
                        {"_register_tools_": cls},
                        word_count=_None,
                    ),
                )
            elif issubclass(subtype, RegisterObject):
                setattr(
                    cls,
                    name,
                    type(
                        subtype.__name__,
                        (subtype,),
                        {"_register_tools_": cls},
                    ),
                )

        cls.AddrMap = type("AddrMap", (AddrMap, cls.RegFile[0]), {})


class RegisterObject(std.Template[GenericArg]):
    # _register_tools_: type[RegisterTools] = None
    _generic_arg_: GenericArg
    _global_offset_: int
    _parent_offset_: GenericArg.offset

    _readable_: bool = True
    _writable_: bool = True

    _cohdlstd_objhasconfig: bool = False

    @_intrinsic
    def _global_word_offset_(self):
        offset, stride = self._global_offset_, self._register_tools_._word_stride_
        assert (
            offset % stride == 0
        ), f"RegisterObject of type {self} is not word aligned"
        return offset // stride

    @classmethod
    def _template_specialize_(cls):
        assert (
            cls._parent_offset_ % cls._register_tools_._word_stride_ == 0
        ), "all RegisterObjects must be word aligned"

    def __init__(self, parent: RegFile, name: str, _cohdlstd_initguard=None):
        self._name_ = name
        self._global_offset_ = parent._global_offset_ + type(self)._parent_offset_

        assert (
            self._global_offset_ % self._register_tools_._word_stride_ == 0
        ), "all RegisterObjects must be word aligned"

    def _impl_print(self) -> TextBlock:
        import inspect

        doc = inspect.getdoc(type(self))

        return TextBlock(
            [] if doc is None else [doc],
            title=f"{self._name_} @ 0x{self._global_offset_:08x}[{self._word_count_}]",
        )

    def _print_(self, *args, **kwargs):
        print(self._impl_print().dump(), *args, **kwargs)

    def _impl_flatten(self, include_devices) -> list[RegisterObject] | RegisterObject:
        return self

    def _flatten_(self, include_devices=False) -> list[RegisterObject]:
        object_list: list[RegisterObject] = _expand_lists(
            self._impl_flatten(include_devices=include_devices)
        )

        assert all(isinstance(obj, RegisterObject) for obj in object_list)

        reg_tools = self._register_tools_

        assert all(obj._register_tools_ is reg_tools for obj in object_list)

        object_list = sorted(object_list, key=lambda obj: obj._global_word_offset_())

        if not include_devices:
            for current, next in zip(object_list, object_list[1:]):
                assert current._global_word_offset_() < next._global_word_offset_()
                assert (
                    current._global_word_offset_() + current._word_count_
                    <= next._global_word_offset_()
                )

        return object_list

    @classmethod
    @_intrinsic
    def _unit_count_(cls):
        return cls._word_count_ * cls._register_tools_._word_stride_

    @classmethod
    @_intrinsic
    def _word_width_(cls):
        return cls._register_tools_._word_width_

    def _contains_addr_(self, addr: Unsigned):
        global_offset = self._global_offset_
        unit_count = self._unit_count_()

        if std.is_pow_two(unit_count) and global_offset % unit_count == 0:
            return addr.msb(rest=std.int_log_2(unit_count)).unsigned == (
                global_offset // unit_count
            )
        else:
            return global_offset <= addr < (global_offset + unit_count)

    def _basic_read_(self, addr, meta):
        return Null

    def _basic_write_(self, addr, data, mask, meta):
        return None

    def _info_(self):
        pass

    def __init_subclass__(cls, readonly=False, writeonly=False):
        if readonly:
            cls._writable_ = False
        if writeonly:
            cls._readable_ = False

        if "_config_" in cls.__dict__:
            cls._cohdlstd_wrappedconfig = cls.__dict__["_config_"]
            cls._config_ = _config_wrapper

        assert (
            cls._writable_ or cls._readable_
        ), "readonly and writeonly cannot be set at the same time"

    def __str__(self):
        return f"{self._name_} : 0x{self._global_offset_:08x} : {type(self).__name__}"


class RegFile(RegisterObject):
    # _member_types_: dict
    _metadata_ = {}

    def __init__(
        self,
        parent: RegFile | None,
        name: str,
        args=None,
        kwargs=None,
        _cohdlstd_initguard=None,
    ):
        if isinstance(self, AddrMap):
            assert (
                parent is None
            ), "AddrMap may only be used as the root instance (use RegFile for nested register collections)"
        else:
            assert (
                parent is not None
            ), "RegFile cannot be used as the root instance (use AddrMap instead)"
            assert (
                args is None and kwargs is None
            ), "RegFile takes no positional or keyword arguments"

        self._name_ = name

        if parent is None:
            self._global_offset_ = 0
        else:
            self._global_offset_ = parent._global_offset_ + type(self)._parent_offset_

        unit_size = self._unit_count_()

        for name, member_type in type(self)._member_types_.items():

            assert (
                member_type._unit_count_() + member_type._parent_offset_ <= unit_size
            ), f"element '{name}' outside the range of its parent device"

            setattr(
                self,
                name,
                member_type(parent=self, name=name, _cohdlstd_initguard=True),
            )

        if isinstance(self, AddrMap):
            if hasattr(self, "_config_"):
                self._config_(*args, **kwargs)

            for obj in self._flatten_(include_devices=True):
                if (
                    hasattr(obj, "_config_")
                    and not "_cohdlstd_objconfigured" in obj.__dict__
                ):
                    obj._config_()

    def _impl_print(self) -> TextBlock:
        block = super()._impl_print()

        for member in type(self)._member_types_:
            block.add(getattr(self, member)._impl_print())

        return block

    def _impl_flatten(self, include_devices) -> list[RegisterObject]:
        if include_devices:
            result = [self]
        else:
            result = []

        result += [
            getattr(self, name)._impl_flatten(include_devices=include_devices)
            for name in type(self)._member_types_
        ]

        return result

    def _basic_read_(self, addr, meta):
        # RegFile is a collection of RegisterObjects
        # _basic_read_ should be called on them, not on the device
        raise AssertionError("_basic_read_ called on RegisterDevice")

    def _basic_write_(self, addr, data, mask, meta):
        # RegFile is a collection of RegisterObjects
        # _basic_write_ should be called on them, not on the device
        raise AssertionError("_basic_write_ called on RegisterDevice")

    def __init_subclass__(cls, *, word_count=_None, readonly=False, writeonly=False):
        if word_count is _None:
            return

        assert cls.__new__ is RegFile.__new__, "RegFile.__new__ cannot be overloaded"
        assert (
            cls.__init__ is RegFile.__init__ or cls.__init__ is AddrMap.__init__
        ), "RegFile.__init__ cannot be overloaded"
        assert (
            cls._basic_read_ is RegFile._basic_read_
        ), "RegFile._basic_read_ cannot be overloaded"
        assert (
            cls._basic_write_ is RegFile._basic_write_
        ), "RegFile._basic_read_ cannot be overloaded"

        super().__init_subclass__(readonly=readonly, writeonly=writeonly)
        cls._word_count_ = word_count

        if hasattr(cls, "_member_types_"):
            members = {**cls._member_types_}
        else:
            members = {}

        metadata = {}

        for name, member_type in get_type_hints(cls, include_extras=True).items():
            ann_data = None

            if hasattr(member_type, "__metadata__"):
                ann_data = member_type.__metadata__
                member_type = member_type.__origin__

            if issubclass(member_type, RegisterObject):
                assert not name in members, f"{name} shadows existing member"

                if ann_data is not None:
                    metadata[name] = ann_data

                for existing_name, existing_type in members.items():
                    if existing_type._parent_offset_ <= member_type._parent_offset_:
                        assert (
                            existing_type._parent_offset_ + existing_type._unit_count_()
                            <= member_type._parent_offset_
                        ), f"member '{name}' (0x{member_type._parent_offset_}-0x{member_type._parent_offset_+member_type._unit_count_()}) overlaps with member '{existing_name}' (0x{existing_type._parent_offset_}-0x{existing_type._parent_offset_+existing_type._unit_count_()}) of type {cls.__name__}"
                    else:
                        assert (
                            member_type._parent_offset_ + member_type._unit_count_()
                            <= existing_name._parent_offset_
                        ), f"member '{name}' (0x{member_type._parent_offset_}-0x{member_type._parent_offset_+member_type._unit_count_()}) overlaps with member '{existing_name}' (0x{existing_type._parent_offset_}-0x{existing_type._parent_offset_+existing_type._unit_count_()}) of type {cls.__name__}"

                members[name] = member_type

        cls._member_types_ = members
        cls._metadata_ = metadata


class AddrMap(RegFile):
    def __init_subclass__(cls, *, word_count=_None, readonly=False, writeonly=False):
        if word_count is _None:
            word_count = _Infinite()
        super().__init_subclass__(
            word_count=word_count, readonly=readonly, writeonly=writeonly
        )

    def _implement_synthesizable_contexts_(self, ctx: SequentialContext):
        content = self._flatten_(include_devices=True)

        for elem in content:
            if hasattr(elem, "_impl_"):
                elem._impl_(ctx)

            if hasattr(elem, "_impl_sequential_"):
                ctx(elem._impl_sequential_)

            if hasattr(elem, "_impl_concurrent_"):
                concurrent(elem._impl_concurrent_)

    def __init__(self, *args, **kwargs):
        super().__init__(
            parent=None, name=type(self).__name__, args=args, kwargs=kwargs
        )


class GenericRegister(RegisterObject):
    _word_count_ = 1

    async def _basic_read_(self, addr, meta):
        return await std.as_awaitable(self._on_read_)

    async def _basic_write_(self, addr, data, mask, meta):
        return await std.as_awaitable(self._on_write_, data, mask)

    def _on_read_(self):
        return Null

    def _on_write_(self, data, mask):
        pass


class Word(GenericRegister):
    _cohdlstd_vector_type = BitVector

    def _config_(self, default=Null):
        self.raw = Signal[
            self._cohdlstd_vector_type[self._register_tools_._word_width_ - 1 : 0]
        ](default)

    def _on_read_(self):
        return self.raw

    @expr_fn
    def __bool__(self):
        return bool(self.raw)

    def __ilshift__(self, src):
        if isinstance(src, Word):
            self.raw <<= src.raw
        else:
            self.raw <<= src
        return self

    @property
    def next(self):
        raise AssertionError("read from next property not supported")

    def next(self, src):
        self <<= src

    def val(self):
        return self.raw


class UWord(Word):
    _cohdlstd_vector_type = Unsigned


class SWord(Word):
    _cohdlstd_vector_type = Signed


class MemWord(Word):
    def _on_write_(self, data, mask: std.Mask):
        self.raw <<= mask.apply(self.raw, data)


class MemUWord(MemWord):
    _cohdlstd_vector_type = Unsigned


class MemSWord(MemWord):
    _cohdlstd_vector_type = Signed


#
#
#


class _NotifyOnRead: ...


class _NotifyOnWrite: ...


class MetaNotifyBase(type):
    @property
    def Read(cls):
        assert not hasattr(cls, "_cohdlstd_onread")
        return cls[_NotifyOnRead]

    @property
    def Write(cls):
        assert not hasattr(cls, "_cohdlstd_onread")
        return cls[_NotifyOnWrite]


class NotifyArg:
    def __init__(self, arg):
        assert (
            arg is _NotifyOnRead or arg is _NotifyOnWrite
        ), "use NotifyType.Read of NotifyType.Write instead of NotifyType[...]"
        self.arg = arg

    def __hash__(self):
        return hash(self.arg)

    def __eq__(self, other):
        return self.arg is other.arg

    def __str__(self):
        if self.arg is _NotifyOnRead:
            return "on_read"
        return "on_write"


class NotifyBase(std.Template[NotifyArg], metaclass=MetaNotifyBase):
    _cohdlstd_notify_mode: NotifyArg.arg

    def notify(self):
        raise AssertionError("abstract method called")

    def __bool__(self):
        raise AssertionError("abstract method called")


class PushOnNotify(NotifyBase):
    def __init__(self):
        self._bit = Signal[Bit](False)

    def notify(self):
        self._bit ^= True

    @expr_fn
    def __bool__(self):
        return self._bit


class FlagOnNotify(NotifyBase):
    def __init__(self):
        self._flag = std.SyncFlag()

    def notify(self):
        self._flag.set()

    @expr_fn
    def __bool__(self):
        return self._flag.is_set()

    def clear(self):
        self._flag.clear()

    async def __aenter__(self):
        await self._flag.__aenter__()

    async def __aexit__(self, val, type, traceback):
        await self._flag.__aexit__(val, type, traceback)


def _primitive_as_int(input):
    if isinstance(input, Bit):
        return 1 if input else 0

    return input.unsigned.to_int()


class _FieldArg:
    offset: int
    width: int
    is_bit: bool
    default: None
    underlying: None

    def __init__(self, arg):
        self.underlying = None
        self.default = None
        underlying_bit = False
        underlying_width = None

        if isinstance(arg, tuple):
            arg, *rest = arg
            if len(rest) == 0:
                pass
            elif len(rest) == 1:
                if isinstance(rest[0], type):
                    self.underlying = rest[0]
                else:
                    self.default = rest[0]
            elif len(rest) == 2:
                assert isinstance(
                    rest[0], type
                ), f"expected second argument to define the field type but got {rest[0]}"
                self.underlying, self.default = rest
            else:
                raise AssertionError(f"to many arguments for field initializer")

        if self.underlying is not None:
            if self.underlying is Bit or (
                issubclass(self.underlying, StdEnum)
                and self.underlying._underlying_ is Bit
            ):
                underlying_bit = True
                underlying_width = 1
            else:
                underlying_width = std.count_bits(self.underlying)

        if isinstance(arg, int):
            self.offset = arg

            if self.underlying is not None:
                self.width = underlying_width
                self.is_bit = underlying_bit
            else:
                self.width = 1
                self.is_bit = True
        else:
            assert isinstance(
                arg, slice
            ), "first generic argument must be integer or slice"
            assert isinstance(arg.start, int), "slice start must be an integer"
            assert isinstance(arg.stop, int), "slice stop must be an integer"
            assert (
                arg.start >= arg.stop
            ), "slice start must be larger than or equal to slice stop"
            assert arg.step is None, "slice step may not be used"

            self.offset = arg.stop
            self.width = arg.start - arg.stop + 1
            self.is_bit = False

            if self.underlying is not None:
                assert (
                    not underlying_bit
                ), "slice argument cannot be used for a Bit type"
                assert (
                    self.width == underlying_width
                ), f"width of slice argument does not match width of underlying type ({underlying_width})"

        if self.underlying is not None:
            if not issubclass(self.underlying, StdEnum):
                if self.is_bit:
                    assert issubclass(self.underlying, Bit)
                else:
                    assert issubclass(self.underlying, BitVector[self.width])

    def __str__(self) -> str:
        if self.is_bit:
            return str(self.offset)
        else:
            return f"{self.offset+self.width-1}:{self.offset}"

    def __hash__(self) -> int:
        return hash(
            (self.offset, self.width, self.is_bit, self.underlying, type(self.default))
        )

    def is_enum(self):
        return self.underlying is not None and issubclass(self.underlying, StdEnum)

    def get_default_as_int(self):
        if self.underlying is None or self.default is None:
            return None

        default_val = self.underlying(self.default)

        if self.is_enum():
            default_val = default_val.raw

        return _primitive_as_int(default_val)

    def __eq__(self, value: _FieldArg) -> bool:
        return (
            self.offset == value.offset
            and self.width == value.width
            and self.is_bit == value.is_bit
            and self.underlying is value.underlying
            and self.default == value.default
        )


class _FlagArg(_FieldArg):
    def __init__(self, arg):
        super().__init__(arg)
        assert (
            self.is_bit
        ), "FlagField argument must be defined as a single bit (not as a slice)"


class _FieldInfoArg:
    args: tuple

    def __init__(self, arg):
        if isinstance(arg, tuple):
            self.args = arg
        else:
            self.args = (arg,)

    def __hash__(self):
        return hash(self.args)

    def __eq__(self, other: tuple):
        return self.args == other.args


class FieldBase:
    _meta_args_ = ()
    pass


class Field(std.Template[_FieldArg], FieldBase):
    _field_arg: _FieldArg

    @classmethod
    @_intrinsic
    def _cohdlstd_underlying(cls):
        arg = cls._field_arg
        if issubclass(cls, UField):
            vec_type = Unsigned

            if arg.underlying is not None:
                assert issubclass(
                    arg.underlying, Unsigned
                ), "explicit underlying type of UField is not unsigned"
        elif issubclass(cls, SField):
            vec_type = Signed

            if arg.underlying is not None:
                assert issubclass(
                    arg.underlying, Signed
                ), "explicit underlying type of SField is not signed"
        else:
            vec_type = BitVector

        if arg.underlying is None:
            if arg.is_bit:
                return Bit
            else:
                return vec_type[arg.width]

        return arg.underlying

    @classmethod
    def _count_bits_(cls):
        return cls._field_arg.width

    @classmethod
    def _from_bits_(cls, val, qualifier):
        return qualifier[cls](val)

    def _to_bits_(self):
        return to_bits(self._value)

    def __ilshift__(self, other):
        if isinstance(other, Field):
            self._value <<= other._value
        else:
            self._value <<= other
        return self

    def __ixor__(self, other):
        if isinstance(other, Field):
            self._value ^= other._value
        else:
            self._value ^= other
        return self

    def __init__(self, value=None, _qualifier_=Signal, extract=False):
        arg = self._field_arg
        underlying = self._cohdlstd_underlying()

        if extract:
            if underlying is Bit:
                self._value = _qualifier_[Bit](value[arg.offset])
            elif issubclass(underlying, BitVector):
                self._value = _qualifier_[underlying[arg.width]](
                    value[arg.offset + arg.width - 1 : arg.offset]
                )
            else:
                assert issubclass(
                    underlying, StdEnum
                ), "underlying field type must be Bit, BitVector or a std.Enum"

                enum_underlying = underlying._underlying_

                assert issubclass(
                    enum_underlying, (Bit, BitVector)
                ), "the underlying type of an enum field must be a Bit or a BitVector"

                if enum_underlying is Bit:
                    self._value = underlying._unsafe_init_(
                        value[arg.offset], _qualifier_
                    )
                else:
                    self._value = underlying._unsafe_init_(
                        value[arg.offset + arg.width - 1 : arg.offset], _qualifier_
                    )

        else:
            if isinstance(value, Field):
                assert (
                    value._field_arg.width == self._field_arg.width
                ), "width of source field does not match target"
                assert (
                    value._field_arg.is_bit == self._field_arg.is_bit
                ), "only one of source and target field is a Bit type"
                other_value = value._value
            else:
                if value is None:
                    other_value = arg.default
                else:
                    other_value = value

            if underlying is Bit:
                self._value = _qualifier_[Bit](other_value)
            elif issubclass(underlying, BitVector):
                self._value = _qualifier_[underlying[arg.width]](other_value)
            else:
                assert issubclass(
                    underlying, StdEnum
                ), "underlying field type must be Bit, BitVector or a std.Enum"

                self._value = _qualifier_[underlying](other_value)

    def val(self):
        return self._value

    @expr_fn
    def __bool__(self):
        return bool(self._value)


class UField(Field):
    pass


class SField(Field):
    pass


class MemField(Field): ...


class MemUField(UField): ...


class MemSField(SField): ...


class FlagField(std.Template[_FlagArg], FieldBase):
    _field_arg: _FlagArg

    @classmethod
    def _count_bits_(cls):
        return 1

    @classmethod
    def _from_bits_(cls, val: BitVector, qualifier=std.Value):
        assert val.width == 1, "FlagField._from_bits_ expects a single bit BitVector"
        return cls(qualifier[Bit](val[0]))

    def _to_bits_(self):
        if self._has_flag:
            return Temporary[Bit](self._flag.is_set())
        else:
            return Temporary[Bit](self._val)

    def __init__(
        self, inp: Bit | BitVector = None, extract=False, _qualifier_=Signal
    ) -> None:
        super().__init__()

        if inp is None:
            assert extract is False, "inp must be set when extract is requested"
            self._flag = std.SyncFlag()
            self._has_flag = True
        else:
            self._has_flag = False
            if extract:
                self._val = _qualifier_[Bit](inp[self._field_arg.offset])
            else:
                assert std.instance_check(
                    inp, Bit
                ), "inp must be a Bit value unless extract is requested"
                self._val = _qualifier_[Bit](inp)

    def flag(self):
        assert (
            self._has_flag
        ), "this method cannot be used on FlagFields constructed from a received bit"
        return self._flag

    def set(self):
        assert (
            self._has_flag
        ), "this method cannot be used on FlagFields constructed from a received bit"
        self._flag.set()

    def clear(self):
        assert (
            self._has_flag
        ), "this method cannot be used on FlagFields constructed from a received bit"
        self._flag.clear()

    def is_set(self):
        if self._has_flag:
            return self._flag.is_set()
        else:
            return self._val

    def is_clear(self):
        if self._has_flag:
            return self._flag.is_clear()
        else:
            return self._val

    async def __aenter__(self):
        await self._flag.__aenter__()

    async def __aexit__(self, val, type, traceback):
        await self._flag.__aexit__(val, type, traceback)


class Register(GenericRegister):
    # _field_types_: dict[str, std.bitfield.FieldBit | std.bitfield.FieldBitVector]
    _metadata_ = {}

    @classmethod
    def _template_deduce_(cls, *args, **kwargs):
        return cls[0]

    # from RegFile
    def _init_from_device(self, parent, name):
        super().__init__(parent, name)
        self._notifications_ = {}
        for name, notification_type in self._notification_types_.items():
            setattr(self, name, notification_type())
            self._notifications_[name] = getattr(self, name)

        self._init_from_config()

    # from user code
    def _init_from_config(self, **kwargs):
        self._fields_ = {}
        for name, field_type in self._field_types_.items():
            setattr(self, name, field_type())
            self._fields_[name] = getattr(self, name)

        for name, default_val in kwargs:
            member = getattr(self, name)
            assert isinstance(member, (MemField, MemUField, MemSField))
            member._set_default_(default_val)

    # from members
    @_intrinsic
    def _init_from_members(self, members):
        assert len(members) == len(self._field_types_)
        self._fields_ = {}

        for name, value in members.items():
            assert isinstance(value, self._field_types_[name])
            setattr(self, name, value)
            self._fields_[name] = value

    def _init_from_user(self, raw=None, _qualifier_=std.Ref, **kwargs):
        if raw is not None:
            assert len(kwargs) == 0
            self._fields_ = {}

            if std.instance_check(raw, BitVector):
                assert (
                    raw.width == self._register_tools_._word_width_
                ), "raw BitVector width does not match register word width"

                for name, field_type in self._field_types_.items():
                    setattr(
                        self,
                        name,
                        field_type(raw, _qualifier_=_qualifier_, extract=True),
                    )
                    self._fields_[name] = getattr(self, name)
            else:
                # raw is another instance of this class

                for name, field_type in self._field_types_.items():
                    setattr(self, name, _qualifier_[field_type](getattr(raw, name)))
                    self._fields_[name] = getattr(self, name)
        else:
            self._init_from_members(**kwargs)

    def __init__(
        self, *args, _cohdlstd_initguard=False, _cohdlstd_frommembers=False, **kwargs
    ):
        if _cohdlstd_initguard:
            self._init_from_device(*args, **kwargs)
        elif _cohdlstd_frommembers:
            assert len(args) == 1
            assert len(kwargs) == 0
            self._init_from_members(args[0])
        else:
            self._init_from_user(*args, **kwargs)

    def __call__(self, **kwargs):
        all_fields = {**self._fields_, **kwargs}
        tmp_fields = {
            name: std.Value[self._field_types_[name]](field_val)
            for name, field_val in all_fields.items()
        }

        return type(self)(tmp_fields, _cohdlstd_frommembers=True)

    @classmethod
    def _count_bits_(cls):
        return cls._register_tools_._word_width_

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier=std.Ref):
        return cls(
            {
                name: qualifier[field_type](bits, extract=True)
                for name, field_type in cls._field_types_.items()
            },
            _cohdlstd_frommembers=True,
        )

    def _to_bits_(self):
        return std.concat(
            *[
                (
                    std.zeros(elem)
                    if isinstance(elem, int)
                    else getattr(self, elem)._to_bits_()
                )
                for elem in self._field_layout_
            ][::-1]
        )

    async def _basic_read_(self, addr, meta):
        for value in self._notifications_.values():
            if value._cohdlstd_notify_mode is _NotifyOnRead:
                value.notify()

        return (await std.as_awaitable(self._on_read_))._to_bits_()

    async def _basic_write_(self, addr, data, mask, meta):
        for value in self._notifications_.values():
            if value._cohdlstd_notify_mode is _NotifyOnWrite:
                value.notify()

        result = await std.as_awaitable(self._on_write_, type(self)._from_bits_(data))

        if result is None:
            # check that self contains no memory
            assert not any(
                [
                    isinstance(field, (MemField, MemUField, MemSField, FlagField))
                    for field in self._fields_.values()
                ]
            )
        else:
            for name, field in self._fields_.items():
                if isinstance(field, (MemField, MemUField, MemSField)):
                    field <<= getattr(result, name)
                elif isinstance(field, FlagField):
                    if getattr(result, name).is_set():
                        field.set()

    def _on_read_(self) -> Register:
        return self

    def _on_write_(self, data):
        return data

    def _assign_(self, source, mode: AssignMode) -> None:
        self._val._assign_(source, mode)

    def __ilshift__(self, source):
        self._val <<= source
        return self

    @property
    def next(self):
        raise AssertionError("next only supported in store context")

    @next.setter
    def next(self, value):
        self._val <<= value

    def _config_(self, **kwargs):
        self._init_from_config(**kwargs)

    def _raw_(self):
        return self._val

    def __init_subclass__(cls, readonly=None, writeonly=None):
        super().__init_subclass__(readonly=readonly, writeonly=writeonly)

        fields: list[tuple[str, type[Field]]] = []
        notifications: list[tuple[str, type[NotifyBase]]] = []
        word_width = cls._register_tools_._word_width_
        metadata = {}

        for name, field_type in get_type_hints(cls, include_extras=True).items():
            if hasattr(field_type, "__metadata__"):
                metadata[name] = field_type.__metadata__
                field_type = field_type.__origin__

            if issubclass(field_type, FieldBase):
                fields.append((name, field_type))

            if issubclass(field_type, NotifyBase):
                notifications.append((name, field_type))

        if len(metadata) != 0:
            cls._metadata_ = metadata

        fields = sorted(fields, key=lambda pair: pair[1]._field_arg.offset)

        # ordered list of field names, padding space is marked
        # by integers giving the number of unused bits
        cls._field_layout_: list[str, int] = []

        offset = 0

        for name, field_type in fields:
            field_offset = field_type._field_arg.offset

            if offset != field_offset:
                assert offset < field_offset, "internal error: offset should be sorted"
                cls._field_layout_.append(field_offset - offset)
            cls._field_layout_.append(name)
            offset = field_offset + field_type._field_arg.width
            assert (
                offset <= word_width
            ), f"register field {name} does not fit in a register of size {word_width}"

        if offset < word_width:
            cls._field_layout_.append(word_width - offset)

        cls._field_types_ = {name: field_type for name, field_type in fields}
        cls._notification_types_ = {
            name: notification_type for name, notification_type in notifications
        }


class Input(GenericRegister, readonly=True):
    def _on_read_(self):
        if self._offset == 0:
            with_offset = self._signal.copy()
        else:
            with_offset = self._signal @ std.zeros(self._offset)

        if self._padding == 0:
            with_padding = with_offset
        else:
            with_padding = std.zeros(self._padding) @ with_offset

        return with_padding

    def _config_(self, signal, /, offset=None, padding=None, lsbs=False, msbs=False):
        assert isinstance(signal, Signal[BitVector])
        width = self._register_tools_._word_width_

        if lsbs:
            assert not msbs, "lsb and msb cannot be set at the same time"
            assert offset is None, "lsb and offset cannot be set at the same time"
            assert padding is None, "lsb and padding cannot be set at the same time"

            offset = 0
            padding = width - signal.width
        elif msbs:
            assert offset is None, "msb and offset cannot be set at the same time"
            assert padding is None, "msb and padding cannot be set at the same time"

            offset = width - signal.width
            padding = 0
        else:
            offset = 0 if offset is None else offset
            padding = 0 if padding is None else padding

        assert (
            signal.width + offset + padding == width
        ), f"resulting input width {signal.width + offset + padding} does not match word width {width}"

        self._signal = signal
        self._width = signal.width
        self._offset = offset
        self._padding = padding


class Output(GenericRegister, writeonly=True):
    def _on_write_(self, data, mask):
        if self._offset == 0:
            with_offset = self._signal
        else:
            with_offset = self._signal @ std.zeros(self._offset)

        if self._padding == 0:
            with_padding = with_offset
        else:
            with_padding = std.zeros(self._padding) @ with_offset

        self._signal <<= (
            mask.apply(with_padding, data)
            .lsb(rest=self._padding)
            .msb(rest=self._offset)
        )

    def _config_(self, signal, offset=None, padding=None, lsbs=False, msbs=False):
        assert isinstance(signal, Signal[BitVector])
        width = self._register_tools_._word_width_

        if lsbs:
            assert not msbs, "lsb and msb cannot be set at the same time"
            assert offset is None, "lsb and offset cannot be set at the same time"
            assert padding is None, "lsb and padding cannot be set at the same time"

            offset = 0
            padding = width - signal.width
        elif msbs:
            assert offset is None, "msb and offset cannot be set at the same time"
            assert padding is None, "msb and padding cannot be set at the same time"

            offset = width - signal.width
            padding = 0
        else:
            offset = 0 if offset is None else offset
            padding = 0 if padding is None else padding

        assert (
            signal.width + offset + padding == width
        ), f"resulting input width {signal.width + offset + padding} does not match word width {width}"

        self._signal = signal
        self._width = signal.width
        self._offset = offset
        self._padding = padding


class Array(RegisterObject):
    _generic_arg_ = None

    def _impl_print(self) -> TextBlock:
        block = super()._impl_print()

        for nr, elem in enumerate(self._elements):
            block.add(elem._impl_print())

        return block

    def _impl_flatten(self, include_devices) -> list[RegisterObject] | RegisterObject:
        if include_devices:
            return [self, *self._elements]
        else:
            return [*self._elements]

    def __init_subclass__(cls, readonly=False, writeonly=False):
        super().__init_subclass__(readonly, writeonly)

        if cls._generic_arg_ is not None:
            assert isinstance(cls._generic_arg_, GenericArg)
            assert (
                cls._generic_arg_.array_step is not None
            ), "array must be defined with a generic step argument"
            cls._word_count_ = (
                cls._generic_arg_.end - cls._generic_arg_.offset
            ) // cls._register_tools_._word_stride_

    def __init__(self, parent: RegFile, name: str, _cohdlstd_initguard=None):
        super().__init__(parent, name)

        arg = type(self)._generic_arg_

        start = arg.offset - self._global_offset_
        stop = arg.end - self._global_offset_
        step = arg.array_step

        elements = []
        prev_offset = self._global_offset_

        for offset in range(start, stop, step):
            # update own offset so child objects are initialized
            # with shifted offset
            # self._global_offset_ = offset
            elements.append(
                arg.array_type[offset](
                    self, "array_elem", _cohdlstd_initguard=_cohdlstd_initguard
                )
            )

        self._elements = elements
        self._global_offset_ = prev_offset

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, index: int):
        return self._elements[index]


class AddrRange(RegisterObject):

    def __init_subclass__(cls, word_count=_None, readonly=False, writeonly=False):
        super().__init_subclass__(readonly, writeonly)

        if hasattr(cls, "_generic_arg_"):
            generic_arg = cls._generic_arg_

            assert isinstance(generic_arg.offset, int)
            assert (
                generic_arg.array_step is None
            ), "AddrRange does not support a generic step parameter"
            assert (
                generic_arg.array_type is None
            ), "AddrRange does not take a generic type parameter"

            word_stride = cls._register_tools_._word_stride_

            if generic_arg.end is not None:
                assert (
                    word_count is _None
                ), "word_count already set using generic argument"
                assert (
                    generic_arg.end - generic_arg.offset
                ) % word_stride == 0, "memory size must be a multiple of the word size"
                word_count = (generic_arg.end - generic_arg.offset) // word_stride

        if word_count is not _None:
            assert not hasattr(
                cls, "_word_count_"
            ), "word_count has already been specified"
            cls._word_count_ = word_count

    async def _basic_read_(self, addr, meta):
        if hasattr(self, "_on_read_"):
            assert not hasattr(
                self, "_on_read_relative_"
            ), "only one of _on_read_ and _on_read_relative_ can be specified"
            return await std.as_awaitable(self._on_read_, addr)
        elif hasattr(self, "_on_read_relative_"):
            return await std.as_awaitable(
                self._on_read_relative_, addr - self._global_offset_
            )

        else:
            return Null

    async def _basic_write_(self, addr, data, mask, meta):
        if hasattr(self, "_on_write_"):
            assert not hasattr(
                self, "_on_write_relative_"
            ), "only one of _on_write_ and _on_write_relative_ can be specified"
            return await std.as_awaitable(self._on_write_, addr, data, mask)
        elif hasattr(self, "_on_write_relative_"):
            return await std.as_awaitable(
                self._on_write_relative_,
                addr - self._global_offset_,
                data,
                mask,
            )


@_intrinsic
def _list_rol(inp: list, roll: int):
    return (inp[roll:] + inp[0:roll])[::-1]


class Memory(AddrRange):
    class MaskMode(enum.Enum):
        IMMEDIATE = enum.auto()
        IGNORE = enum.auto()
        READBACK = enum.auto()
        SPLIT_WORDS = enum.auto()

    def _config_(
        self,
        initial=None,
        noreset=False,
        mask_mode: MaskMode = None,
        allow_unaligned=False,
        inline=False,
    ):
        if mask_mode is None:
            mask_mode = Memory.MaskMode.IMMEDIATE

        if allow_unaligned:
            assert (
                mask_mode is Memory.MaskMode.SPLIT_WORDS
            ), f"unaligned access is currently only supported for MaskMode.SPLIT_WORDS"

        self._register_tools_: RegisterTools
        self._inline = inline

        word_width = type(self)._word_width_()
        word_cnt = type(self)._word_count_

        self._word_stride = self._register_tools_._word_stride_
        self._unit_width = self._register_tools_._addr_unit_width_

        self._mask_mode = mask_mode
        self._allow_unaligned = allow_unaligned

        Qualifier = NoresetSignal if noreset else Signal

        if mask_mode is Memory.MaskMode.SPLIT_WORDS:

            def split_initial(nr):
                if initial is None or initial is Null or initial is Full:
                    return initial
                else:
                    right_bit = self._unit_width * nr
                    left_bit = right_bit + self._unit_width - 1

                    return [
                        BitVector[word_width](elem)[left_bit:right_bit]
                        for elem in initial
                    ]

            self._mem_list = [
                Qualifier[std.Array[BitVector[self._unit_width], word_cnt]](
                    split_initial(nr)
                )
                for nr in range(self._word_stride)
            ]
        else:
            self._mem = Qualifier[std.Array[BitVector[word_width], word_cnt]](initial)

    async def _cohdlstd_impl_read(self, addr):
        waddr = self._register_tools_._as_word_addr_(addr)
        if self._mask_mode is self.MaskMode.SPLIT_WORDS:
            stride_cnt = int_log_2(self._word_stride)
            stride_bits = addr.lsb(stride_cnt).unsigned

            if not self._allow_unaligned:
                assert stride_bits == 0, "stride error"

                return concat(
                    *[
                        self._mem_list[unit_nr][waddr]
                        for unit_nr in range(self._word_stride)
                    ][::-1]
                )
            else:

                rddata = [
                    self._mem_list[off][
                        waddr
                        + (Unsigned[1](0) if stride_bits <= off else Unsigned[1](1))
                    ]
                    for off in range(self._word_stride)
                ]

                result = Variable[BitVector[self._word_width_()]]()

                for off in range(self._word_stride):
                    if stride_bits == off:
                        result @= concat(*_list_rol(rddata, off))

                return result
        else:
            return self._mem[waddr]

    async def _cohdlstd_impl_write(self, addr, data, mask):
        waddr = Variable(self._register_tools_._as_word_addr_(addr))

        if self._mask_mode is self.MaskMode.IGNORE:
            self._mem[waddr] <<= data
        elif self._mask_mode is self.MaskMode.IMMEDIATE:
            self._mem[waddr] <<= mask.apply(self._mem[waddr], data)
        elif self._mask_mode is self.MaskMode.READBACK:
            prev_val = Signal(self._mem[waddr])
            await true
            self._mem[waddr] <<= mask.apply(prev_val, data)
        elif self._mask_mode is self.MaskMode.SPLIT_WORDS:
            unit_mask = mask.as_vector(self._word_width_())

            if not self._allow_unaligned:
                for unit_nr in range(self._word_stride):
                    right_bit = self._unit_width * unit_nr
                    left_bit = right_bit + self._unit_width - 1

                    if unit_mask[unit_nr * self._unit_width]:
                        self._mem_list[unit_nr][waddr] <<= data[left_bit:right_bit]
            else:
                stride_cnt = int_log_2(self._word_stride)
                stride_bits = addr.lsb(stride_cnt).unsigned

                addr_list = [
                    Variable[base_type(waddr)]() for _ in range(self._word_stride)
                ]
                data_list = [
                    Variable[BitVector[self._unit_width]]()
                    for _ in range(self._word_stride)
                ]
                mask_list = [Variable[Bit]() for _ in range(self._word_stride)]

                for unit_nr in range(self._word_stride):
                    val = addr_list[unit_nr]

                    val @= waddr + (
                        Unsigned[1](0) if stride_bits <= unit_nr else Unsigned[1](1)
                    )

                for off in range(self._word_stride):
                    if off == stride_bits:
                        for unit_nr in range(self._word_stride):
                            lsb = self._find_lsb_of_unaligned_split(off, unit_nr)

                            msb = lsb + self._unit_width - 1

                            data_list[unit_nr] @= data[msb:lsb]
                            mask_list[unit_nr] @= unit_mask[lsb]

                for unit_nr in range(self._word_stride):
                    if mask_list[unit_nr]:
                        self._mem_list[unit_nr][addr_list[unit_nr]] <<= data_list[
                            unit_nr
                        ]

    def _impl_(self, ctx: SequentialContext):
        if not self._inline:
            self._rd_enable = Signal[Bit](False)
            self._rd_addr = Signal[Unsigned.upto(self._unit_count_())]()
            self._rd_data = Signal[Unsigned[self._word_width_()]]()

            self._wr_enable = Signal[Bit](False)
            self._wr_addr = Signal[Unsigned.upto(self._unit_count_())]()
            self._wr_data = Signal[Unsigned[self._word_width_()]]()

            if self._mask_mode is not self.MaskMode.IGNORE:
                self._wr_mask = Signal[BitVector[self._word_width_()]]()
            else:
                self._wr_mask = None

            @ctx
            async def impl_read_memory():
                if self._rd_enable:
                    self._rd_data <<= await self._cohdlstd_impl_read(self._rd_addr)

            @ctx
            async def impl_write_memory():
                if self._mask_mode is self.MaskMode.READBACK:
                    while True:
                        if self._wr_enable:
                            await self._cohdlstd_impl_write(
                                self._wr_addr, self._wr_data, std.Mask(self._wr_mask)
                            )
                            continue
                else:
                    if self._wr_enable:
                        await self._cohdlstd_impl_write(
                            self._wr_addr, self._wr_data, std.Mask(self._wr_mask)
                        )

    async def _on_read_relative_(self, addr):
        if self._inline:
            return await self._cohdlstd_impl_read(addr)
        else:
            self._rd_addr <<= addr.lsb(self._rd_addr.width)
            self._rd_enable ^= True
            await true
            await true
            return self._rd_data

    async def _on_write_relative_(self, addr, data, mask):
        if self._inline:
            await self._cohdlstd_impl_write(addr, data, mask)
        else:
            if self._mask_mode is not self.MaskMode.IGNORE:
                self._wr_mask <<= mask.as_vector(self._word_width_())

            self._wr_addr <<= addr.lsb(self._wr_addr.width)
            self._wr_data <<= data
            self._wr_enable ^= True

    @_intrinsic
    def _find_lsb_of_unaligned_split(self, addr_offset, unit_nr):
        word_width = self._word_width_()
        unit_width = self._unit_width

        return (word_width + ((unit_nr - addr_offset) * unit_width)) % word_width


class RoMemory(Memory):
    def __init_subclass__(cls, word_count=_None):
        return super().__init_subclass__(word_count, readonly=True, writeonly=False)

    def _config_(self, initial, mask_mode=None, inline=False):
        return super()._cohdlstd_wrappedconfig(
            initial, noreset=True, mask_mode=mask_mode, inline=inline
        )


RegisterTools.RegisterObject = RegisterObject
RegisterTools.RegFile = RegFile
RegisterTools.AddrMap = AddrMap

RegisterTools.NotifyBase = NotifyBase
RegisterTools.PushOnNotify = PushOnNotify
RegisterTools.FlagOnNotify = FlagOnNotify

RegisterTools.GenericRegister = GenericRegister
RegisterTools.Register = Register

RegisterTools.Word = Word
RegisterTools.UWord = UWord
RegisterTools.SWord = SWord
RegisterTools.MemWord = MemWord
RegisterTools.MemUWord = MemUWord
RegisterTools.MemSWord = MemSWord

RegisterTools.Access = Access
RegisterTools.HwAccess = HwAccess
RegisterTools.Field = Field
RegisterTools.UField = UField
RegisterTools.SField = SField
RegisterTools.MemField = MemField
RegisterTools.MemUField = MemUField
RegisterTools.MemSField = MemSField
RegisterTools.FlagField = FlagField


RegisterTools.Input = Input
RegisterTools.Output = Output
RegisterTools.Array = Array
RegisterTools.AddrRange = AddrRange

RegisterTools.Memory = Memory
RegisterTools.RoMemory = RoMemory

reg32 = RegisterTools[32, 8]



from __future__ import annotations

import enum

from ._primitive_type import _PrimitiveType
from ._boolean import _NullFullType
from ._bit_vector import BitVector

class BitState(enum.Enum):
    LOW = enum.auto()
    HIGH = enum.auto()
    UNINITIALZED = enum.auto()
    UNKNOWN = enum.auto()
    HIGH_IMPEDANCE = enum.auto()
    WEAK = enum.auto()
    WEAK_LOW = enum.auto()
    WEAK_HIGH = enum.auto()
    DONT_CARE = enum.auto()

    def __bool__(self) -> bool: ...
    def __invert__(self) -> BitState: ...
    def __or__(self, other) -> BitState: ...
    def __and__(self, other) -> BitState: ...
    def __xor__(self, other) -> BitState: ...
    @staticmethod
    def from_str(char: str) -> BitState: ...
    @staticmethod
    def construct(
        arg: BitState | Bit | str | int | bool | _NullFullType | None,
    ) -> BitState: ...

class Bit(_PrimitiveType):
    def __init__(
        self,
        value: Bit
        | None
        | BitState
        | str
        | int
        | bool
        | _NullFullType = BitState.UNINITIALZED,
    ):
        self._val: BitState
    def copy(self) -> Bit: ...
    @property
    def type(self):
        return Bit
    def get(self) -> BitState: ...
    def invert(self) -> Bit: ...
    def __invert__(self) -> Bit: ...
    def __inv__(self) -> Bit: ...
    def __or__(self, other: Bit) -> Bit: ...
    def __and__(self, other: Bit) -> Bit: ...
    def __xor__(self, other: Bit) -> Bit: ...
    def __matmul__(self, other) -> BitVector: ...
    def __bool__(self) -> bool: ...
    def __eq__(self, other: Bit | BitState | bool | str) -> bool: ...
    def __ne__(self, other: Bit | BitState | bool | str) -> bool: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...



import cohdl
from cohdl.std._context import concurrent, sequential


from .base import Axi4Light

#
#
# interconnect
#
#


def background_range(clk, reset, addr_width):
    axi = Axi4Light.signal(
        clk,
        reset,
        addr_width=addr_width,
        data_width=32,
    )

    @concurrent
    def background_axi():
        axi.rdaddr.ready <<= True

        axi.rddata.valid <<= True
        axi.rddata.rdata <<= cohdl.Null
        axi.rddata.rresp <<= Axi4Light.RespConstants.DECERR

        axi.wraddr.ready <<= True
        axi.wrdata.ready <<= True
        axi.wrresp.valid <<= True
        axi.wrresp.bresp <<= Axi4Light.RespConstants.DECERR

    return axi


class Interconnect:
    class SlaveWrapper:
        def __init__(self, offset: int, size: int, axi: Axi4Light):
            assert size.bit_count() == 1, "size must be a power of 2"
            assert offset % size == 0, "offset must be a multiple of the interface size"

            self.range_start = offset
            self.range_end = offset + size - 1
            self.axi = axi
            self.rd_active = cohdl.Signal[cohdl.Bit](False)
            self.wr_active = cohdl.Signal[cohdl.Bit](False)

        def when_active(self, value):
            return value if self.active else cohdl.Null

        def contains_addr(self, addr):
            return self.range_start <= addr.unsigned <= self.range_end

        def overlapps(self, offset, size):
            return (
                (self.range_start <= offset <= self.range_end)
                or (self.range_start <= offset + size <= self.range_end)
                or (offset <= self.range_start < offset + size)
                or (offset <= self.range_end < offset + size)
            )

    def _all_slaves(self):
        return [*self._slaves, self._background]

    def __init__(self, master: Axi4Light):
        addr_width = master.addr_width()

        self._clk = master.clk
        self._reset = master.reset
        self._background = Interconnect.SlaveWrapper(
            0, 2**addr_width, background_range(self._clk, self._reset, addr_width)
        )
        self._slaves: list[Interconnect.SlaveWrapper] = []

        @sequential
        def proc_connect():
            for slv in self._all_slaves():
                slv_addr_width = slv.axi.addr_width()

                if slv.rd_active:
                    master.rdaddr.ready ^= slv.axi.rdaddr.ready
                    slv.axi.rdaddr.valid ^= master.rdaddr.valid
                    slv.axi.rdaddr.araddr ^= master.rdaddr.araddr.lsb(slv_addr_width)
                    slv.axi.rdaddr.arprot ^= master.rdaddr.arprot

                    slv.axi.rddata.ready ^= master.rddata.ready
                    master.rddata.valid ^= slv.axi.rddata.valid
                    master.rddata.rdata ^= slv.axi.rddata.rdata
                    master.rddata.rresp ^= slv.axi.rddata.rresp

                if slv.wr_active:
                    master.wraddr.ready ^= slv.axi.wraddr.ready
                    slv.axi.wraddr.valid ^= master.wraddr.valid
                    slv.axi.wraddr.awaddr ^= master.wraddr.awaddr.lsb(slv_addr_width)
                    slv.axi.wraddr.awprot ^= master.wraddr.awprot

                    master.wrdata.ready ^= slv.axi.wraddr.ready
                    slv.axi.wrdata.valid ^= master.wrdata.valid
                    slv.axi.wrdata.wdata ^= master.wrdata.wdata
                    slv.axi.wrdata.wstrb ^= master.wrdata.wstrb

                    slv.axi.wrresp.ready ^= master.wrresp.ready
                    master.wrresp.valid ^= slv.axi.wrresp.valid
                    master.wrresp.bresp ^= slv.axi.wrresp.bresp

        @sequential(self._clk, self._reset)
        async def proc_read():
            await master.rdaddr.valid

            for slv in self._all_slaves():
                if slv.contains_addr(master.rdaddr.araddr):
                    slv.rd_active <<= True

            await master.rddata.valid & master.rddata.ready

            for slv in self._all_slaves():
                slv.rd_active <<= False

        @sequential(self._clk, self._reset)
        async def proc_write():
            await master.wraddr.valid

            for slv in self._all_slaves():
                if slv.contains_addr(master.wraddr.awaddr):
                    slv.wr_active <<= True

            await master.wrresp.valid & master.wrresp.ready

            for slv in self._all_slaves():
                slv.wr_active <<= False

    def reserve(self, offset: int, size: int, prefix="") -> Axi4Light:
        assert size.bit_count() == 1, "size must be a power of 2"
        addr_width = size.bit_length() - 1

        axi = Axi4Light.signal(self._clk, self._reset, addr_width, prefix=prefix)

        self.connect(axi, offset, size)

        return axi

    def connect(self, axi: Axi4Light, offset: int, size: int | None = None):
        requested_size = size
        actual_size = 2 ** axi.addr_width()

        assert requested_size is None or (requested_size == actual_size)

        for slv in self._slaves:
            assert not slv.overlapps(offset, actual_size)

        self._slaves.append(Interconnect.SlaveWrapper(offset, actual_size, axi))



from __future__ import annotations

import enum

from cohdl._core._ir import repr as ir
from cohdl._core._type_qualifier import (
    Signal,
    Temporary,
    Port,
    TypeQualifier,
)
from cohdl.utility import IdMap
from cohdl.utility.id_map import IdSet

from cohdl._core._inline import InlineRaw, InlineVhdl
from cohdl._core._intrinsic import _SensitivityAll, _SensitivityList

from . import _vhdl_repr as vhdl


def assign_temporary(target, source: vhdl.Expression, kwargs):
    ctx: Context = kwargs["context"]
    assert isinstance(target, (Temporary, Signal))

    if ctx is Context.CONCURRENT:
        # temporaries in concurrent statements are implemented as Signals
        return vhdl.SignalAssignment(vhdl.Target(target), source)
    else:
        return vhdl.VariableAssignment(vhdl.Target(target), source)


class _StmtAssembler:
    @staticmethod
    def convert_statement(stmt: ir.Statement, *args, **kwargs):
        return _StmtAssembler().apply(stmt, *args, **kwargs)

    #
    #
    #
    #

    def apply(self, inp, *args, **kwargs):
        if isinstance(
            inp, (ir.SignalAssignment | ir.SignalPush | ir.VariableAssignment)
        ):
            ctx: Context = kwargs["context"]

            if isinstance(inp, ir.SignalAssignment):
                is_signal_assign = True
            elif isinstance(inp, ir.SignalPush):
                assert ctx is Context.SEQUENTIAL
                is_signal_assign = True
            elif isinstance(inp, ir.VariableAssignment):
                assert (
                    ctx is Context.SEQUENTIAL
                ), "variable assignment only possible in sequential contexts"
                is_signal_assign = False
            else:
                raise AssertionError()

            if is_signal_assign:
                return vhdl.SignalAssignment(
                    vhdl.Target(inp._target), vhdl.Value(inp._source)
                )
            else:
                return vhdl.VariableAssignment(
                    vhdl.Target(inp._target), vhdl.Value(inp._source)
                )

        if isinstance(inp, ir.Nop):
            return vhdl.Nop()

        if isinstance(inp, ir.Comment):
            return vhdl.Comment(inp.lines)

        if isinstance(inp, ir.BinOp):
            return assign_temporary(
                inp.result(),
                vhdl.BinOp(
                    inp._op,
                    vhdl.Value(inp._lhs),
                    vhdl.Value(inp._rhs),
                    inp.result(),
                ),
                kwargs,
            )

        if isinstance(inp, ir.UnaryOp):
            return assign_temporary(
                inp.result(),
                vhdl.UnaryOp(inp._op, vhdl.Value(inp._arg), inp.result()),
                kwargs,
            )

        if isinstance(inp, ir.Compare):
            return assign_temporary(
                inp.result(),
                vhdl.Compare(
                    inp._op,
                    vhdl.Value(inp._lhs),
                    vhdl.Value(inp._rhs),
                    inp.result(),
                ),
                kwargs,
            )

        if isinstance(inp, ir.Boolean):
            return assign_temporary(
                inp.result(), vhdl.Boolean(vhdl.Value(inp._arg)), kwargs
            )

        if isinstance(inp, ir.All):
            return assign_temporary(
                inp.result(),
                vhdl.All([vhdl.Value(arg) for arg in inp._args], inp.result()),
                kwargs,
            )

        if isinstance(inp, ir.Any):
            return assign_temporary(
                inp.result(),
                vhdl.Any([vhdl.Value(arg) for arg in inp._args], inp.result()),
                kwargs,
            )

        if isinstance(inp, ir.CodeBlock):
            return vhdl.CodeBlock(
                [self.apply(stmt, **kwargs) for stmt in inp.content()]
            )

        if isinstance(inp, ir.If):

            def conv(ev: ir.Event | ir.EventGroup):
                if isinstance(ev, ir.Event):
                    return vhdl.Event(vhdl.Value(ev.sig), ev.event_type)
                else:
                    if ev.operation is ir.EventGroup.Operation.AND:
                        return vhdl.All([conv(e) for e in ev.events], True)
                    else:
                        return vhdl.Any([conv(e) for e in ev.events], True)

            if isinstance(inp._test, (ir.Event, ir.EventGroup)):
                test = conv(inp._test)
            else:
                test = vhdl.Boolean(vhdl.Value(inp._test))

            return vhdl.If(
                test,
                self.apply(inp._body, **kwargs),
                self.apply(inp._orelse, **kwargs),
            )

        if isinstance(inp, ir.SelectWith):
            if inp._default is None:
                default = None
            else:
                default = vhdl.Value(inp._default)

            use_case_when = kwargs["context"] is Context.SEQUENTIAL

            if use_case_when:
                result = inp.result()

                if isinstance(result, Signal):
                    Assignment = vhdl.SignalAssignment
                else:
                    Assignment = vhdl.VariableAssignment

                if default is not None:
                    default = vhdl.CodeBlock([Assignment(vhdl.Target(result), default)])

                return vhdl.CaseWhen(
                    vhdl.Value(inp._arg),
                    [
                        (
                            vhdl.Constant(branch[0]),
                            vhdl.CodeBlock(
                                [Assignment(vhdl.Target(result), vhdl.Value(branch[1]))]
                            ),
                        )
                        for branch in inp._branches
                    ],
                    default,
                )

            return vhdl.SelectWith(
                vhdl.Value(inp._arg),
                [
                    (vhdl.Constant(branch[0]), vhdl.Value(branch[1]))
                    for branch in inp._branches
                ],
                default,
                vhdl.Target(inp.result()),
            )

        if isinstance(inp, ir.CaseWhen):
            branches = []

            for branch in inp._branches:
                cond = vhdl.Constant(branch.cond)
                code = self.apply(branch.code, **kwargs)

                branches.append((cond, code))

            if inp._default is None:
                default = None
            else:
                default = self.apply(inp._default, **kwargs)

            return vhdl.CaseWhen(vhdl.Value(inp._value), branches, default)

        if isinstance(inp, ir.Assert):
            return vhdl.Assert(vhdl.Boolean(vhdl.Value(inp._cond)), inp._msg)

        if isinstance(inp, ir.InlineCode):
            for option in inp.options:
                if isinstance(option, (InlineRaw, InlineVhdl)):
                    if inp.result is not None:
                        return assign_temporary(
                            inp.result, vhdl.InlineCode(option, inp.result), kwargs
                        )
                    else:
                        return vhdl.InlineCode(option, inp.result)

            raise AssertionError("no vhdl option for the inline code exists")

        raise AssertionError(f"cannot convert {inp}")


#
#
#
# instances
#
#
#


class Context(enum.Enum):
    SEQUENTIAL = enum.auto()
    CONCURRENT = enum.auto()


class VhdlAssembler:
    def __init__(self):
        self._known_templates: IdMap[ir.EntityTemplate, vhdl.Entity] = IdMap()
        self._stmt_assembler = _StmtAssembler()

    def convert_stmt(self, stmt, **kwargs):
        return self._stmt_assembler.apply(stmt, **kwargs)

    def _get_known_templates(self):
        return self._known_templates

    def _add_template(self, inp, ret):
        assert inp not in self._known_templates
        self._known_templates[inp] = ret

    #
    #
    #
    #

    def apply(self, inp, **kwargs):
        if isinstance(inp, ir.Entity):
            scope = kwargs["parent_scope"]

            for val in inp.get_ports().values():
                scope.declare(val._root)

            # call convert template to add it to
            # the list of vhdl Entities
            entity = self.apply(inp.get_template(), scope=scope)

            return vhdl.EntityInst(scope, entity, inp.get_ports(), inp.get_generics())

        if isinstance(inp, ir.EntityTemplate):
            if inp in self._get_known_templates():
                return self._get_known_templates()[inp]

            module_scope = vhdl.ModuleScope()
            entity_scope = vhdl.EntityScope(module_scope)
            arch_scope = vhdl.ArchScope(entity_scope)

            output_ports = []

            for name, port in inp.port_declarations().items():
                entity_scope.declare(port, name_hint=name)

                if port.direction() == Port.Direction.OUTPUT:
                    output_ports.append(port)

            for name, generic in inp.generic_declarations().items():
                entity_scope.declare(generic, name_hint=name)

            if "reserved_names" in inp._attributes:
                reserved_names = inp._attributes["reserved_names"]

                for name in reserved_names:
                    arch_scope.reserve_name(name)

            blocks = []

            # search buffer ports
            buffer_ports = output_ports
            # create buffer signals
            buffer_assignments = []

            alias_scope = vhdl.AliasScope(arch_scope)

            for port in buffer_ports:
                if port.has_default():
                    buffer = Signal[port.type](
                        port.default(),
                        name=(
                            f"buffer{port.name()}"
                            if port.name().endswith("_")
                            else f"buffer_{port.name()}"
                        ).strip("_"),
                    )
                else:
                    buffer = Signal[port.type](
                        name=(
                            f"buffer{port.name()}"
                            if port.name().endswith("_")
                            else f"buffer_{port.name()}"
                        ).strip("_")
                    )

                buffer_assignments.append(
                    vhdl.SignalAssignment(vhdl.Target(port), vhdl.Value(buffer))
                )

                arch_scope.declare(buffer)
                alias_scope.set_alias(port, buffer)

            # assign buffer to port
            # use buffer instead of port

            blocks.append(
                vhdl.Concurrent(arch_scope, buffer_assignments, "buffer assignment", {})
            )

            # convert subblocks
            for block in inp.subblocks():
                blocks.append(
                    self.apply(block, **{**kwargs, "parent_scope": alias_scope})
                )

            for ctx in inp.contexts():
                blocks.append(
                    self.apply(ctx, **{**kwargs, "parent_scope": alias_scope})
                )

            ret = vhdl.Entity(inp.info(), alias_scope, blocks)  # type: ignore

            arch = vhdl.Architecture(alias_scope, ret, blocks)
            entity_scope.declare(arch)
            ret._arch = arch

            self._add_template(inp, ret)

            module_scope.complete_setup()

            return ret

        if isinstance(inp, ir.Block):
            return vhdl.Block(
                kwargs["parent_scope"],
                [self.apply(subblock, **kwargs) for subblock in inp.subblocks()]
                + [self.apply(ctx, **kwargs) for ctx in inp.contexts()],
                name=inp.name(),
                attributes=inp._attributes,
            )

        if isinstance(inp, ir.Concurrent):
            parent_scope: vhdl.VhdlScope = kwargs["parent_scope"]

            def collect_objects(obj, access):
                if isinstance(obj, TypeQualifier):
                    parent_scope.declare(obj._root)
                return obj

            inp.code().visit_referenced_objects(collect_objects)

            return vhdl.Concurrent(
                parent_scope,
                [
                    _StmtAssembler.convert_statement(
                        stmt, **{**kwargs, "context": Context.CONCURRENT}
                    )
                    for stmt in inp.code().content()
                ],
                inp.name(),
                inp.attributes,
            )

        if isinstance(inp, ir.Sequential):
            parent_scope: vhdl.VhdlScope = kwargs["parent_scope"]
            scope = vhdl.ProcessScope(parent_scope)

            def collect_objects(obj, access):
                if isinstance(obj, TypeQualifier):
                    scope.declare(obj._root)
                return obj

            inp.visit_referenced_objects(collect_objects)

            kwargs = {**kwargs, "context": Context.SEQUENTIAL}

            code = vhdl.CodeBlock(
                [self.convert_stmt(inp._code, context=Context.SEQUENTIAL)]
            )

            #
            #
            #

            proc_name = inp.name()

            if proc_name is None:
                proc_name = "process"

            # since process(all) is not supported prior to VHDL-2008
            # instead search for all signals, that are read in the process
            # and specify them explicitly in the sensitivity list
            if isinstance(inp._sensitivity, _SensitivityAll):
                read_roots = IdSet()

                def find_read_roots(obj, access: ir.AccessFlags):
                    if access is ir.AccessFlags.READ and isinstance(obj, Signal):
                        read_roots.add(obj._root)
                    return obj

                inp.visit_referenced_objects(find_read_roots)

                sensitivity = _SensitivityList(read_roots)
            else:
                sensitivity = inp._sensitivity

            proc = vhdl.Process(scope, code, sensitivity, inp.attributes)
            parent_scope.declare(proc, name_hint=proc_name)

            if inp._always_expr is not None:
                concurrent = vhdl.Concurrent(
                    scope,
                    [
                        self.convert_stmt(
                            inp._always_expr.code(), context=Context.CONCURRENT
                        )
                    ],
                    f"always - {proc_name}",
                    {},
                )

                return vhdl.Block(
                    scope,
                    [
                        concurrent,
                        proc,
                    ],
                    None,
                    {},
                )

            return proc


class VhdlMakeLibrary:
    def apply(self, inp):
        assert isinstance(inp, vhdl.Instance)
        return vhdl.Library.from_top_entity(inp)



from cohdl import Bit, Signal
from cohdl import vhdl as raw_vhdl

from cohdl import pyeval, Attribute

import enum


class CmpType(enum.Enum):
    ID = enum.auto()
    EQ = enum.auto()


class Cache:
    def __init__(self, arg_types):
        self._arg_types = arg_types
        self._entries = []

    @pyeval
    def lookup(self, args):
        assert len(args) == len(self._arg_types)

        for cached_value, cached_args in self._entries:
            for cmp_type, arg, cached in zip(self._arg_types, args, cached_args):
                if cmp_type is CmpType.ID and arg is not cached:
                    break
                elif arg != cached:
                    break
            else:
                return (True, cached_value)

        return (False, None)

    @pyeval
    def add_entry(self, value, args):
        assert len(args) == len(self._arg_types)
        self._entries.append((value, args))


@pyeval
def cached(*cmp_types):
    cache = Cache(cmp_types)

    def helper(fn):
        def wrapped(*args):
            found, value = cache.lookup(args)

            if found:
                return value
            else:
                new_entry = fn(*args)
                cache.add_entry(new_entry, args)
                return new_entry

        return wrapped

    return helper


class vhdl(raw_vhdl):
    def post_process(self, text: str):
        result = text.replace("<%", "{")
        return result.replace("%>", "}")


#
# attributes
#


class anyconst(Attribute, type=bool): ...


@cached(CmpType.ID)
def stable(arg):
    return f"{vhdl[bool]:stable({arg!r})}"


@cached(CmpType.ID, CmpType.EQ)
def _cached_prev(sig, cnt):
    if cnt is None:
        return f"{vhdl[sig.type]:prev({sig!r})}"
    else:
        return f"{vhdl[sig.type]:prev({sig!r}, {cnt})}"


def prev(sig, cnt=None):
    return _cached_prev(sig, cnt)


#
#
#


def _n_prev_true(sig, n: int):
    if n == 0:
        return sig
    else:
        return sig and _n_prev_true(prev(sig), n - 1)


def n_prev_true(sig, n: int | tuple):
    if isinstance(n, int):
        return _n_prev_true(sig, n)
    else:
        min_n, max_n = n
        return _n_prev_true(sig, min_n) and not prev(sig, max_n + 1)



from __future__ import annotations

from cohdl._core import Bit, BitVector, Signal, Unsigned, Null, Full, TypeQualifierBase

from ._context import Context, Duration, Frequency, concurrent_eval, concurrent_assign
from ._core_utility import instance_check, is_qualified, as_bitvector


from .utility import (
    InShiftRegister,
    OutShiftRegister,
    ToggleSignal,
    max_int,
)

from ._prefix import prefix


class SpiMode:
    def __init__(self, cpol: bool, cpha: bool):
        self.cpol = cpol
        self.cpha = cpha

    @staticmethod
    def mode(val: int):
        assert val in (0, 1, 2, 3)
        return SpiMode(cpol=val // 2, cpha=val % 2)


class Spi:
    def __init__(
        self,
        sclk: Signal[Bit],
        mosi: Signal[Bit] | Signal[BitVector],
        miso: Signal[Bit] | Signal[BitVector],
        chip_select: Signal[Bit] | Signal[BitVector],
        mode: SpiMode | None = None,
    ):
        if mode is None:
            mode = SpiMode.mode(0)

        self.sclk = sclk
        self.mosi = mosi
        self.miso = miso
        self.chip_select = chip_select
        self.mode = mode


class _SpiTransaction:
    def __init__(self, master: SpiMaster):
        self._master = master

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._master._end_transaction()

    async def send_data(self, data: BitVector | str):
        await self._master._send_data(data, _no_wait=True)

    async def read_data(self, len: int):
        return await self._master._read_data(len)


class SpiMaster:
    def __init__(
        self,
        ctx: Context,
        spi: Spi,
        *,
        clk_period: Unsigned | int | Duration | None = None,
        clk_frequency: Frequency | None = None,
    ):
        assert (
            clk_period is None or clk_frequency is None
        ), "spi clock period or frequency must be set"

        if clk_period is None:
            clk_period = clk_frequency.period()

        self._prefix = prefix("spi")

        with self._prefix:
            if is_qualified(clk_period):
                assert instance_check(clk_period, Unsigned)
                half_period = Signal[Unsigned.upto(max_int(clk_period))]()
                concurrent_eval(half_period, lambda: clk_period >> 1)
            else:
                if isinstance(clk_period, Duration):
                    half_period = clk_period / 2
                else:
                    half_period = clk_period // 2

            self.spi = spi

            self._toggle = ToggleSignal(
                ctx,
                half_period,
                require_enable=True,
                default_state=spi.mode.cpol,
                first_state=spi.mode.cpol ^ spi.mode.cpha,
            )

            self._cs = Signal[type(TypeQualifierBase.decay(self.spi.chip_select))](Full)

            concurrent_eval(self.spi.sclk, self._toggle.state)
            concurrent_assign(self.spi.chip_select, self._cs)

    def _shift_edge(self):
        if self.spi.mode.cpol == self.spi.mode.cpha:
            return self._toggle.falling()
        else:
            return self._toggle.rising()

    def _sample_edge(self):
        if self.spi.mode.cpol == self.spi.mode.cpha:
            return self._toggle.rising()
        else:
            return self._toggle.falling()

    def _start_transaction(self, cs: Bit | BitVector | None):
        if cs is None:
            assert instance_check(self._cs, Bit)
            self._cs <<= False
        else:
            self._cs <<= cs

        self._toggle.enable()

    def _end_transaction(self):
        self._toggle.disable()
        self._cs <<= Full

    async def _send_data(self, data: BitVector, *, _no_wait=False):
        data_tx = OutShiftRegister(data, msb_first=True)

        shift_width = (
            None if instance_check(self.spi.mosi, Bit) else self.spi.mosi.width
        )

        if not _no_wait and self.spi.mode.cpha:
            await self._shift_edge()

        self.spi.mosi <<= data_tx.shift(shift_width)

        while not data_tx.empty():
            await self._shift_edge()
            self.spi.mosi <<= data_tx.shift(shift_width)

        await self._shift_edge()

        self.spi.mosi <<= Null

    async def _read_data(self, len: int):
        data_rx = InShiftRegister(len, msb_first=True)

        while not data_rx.full():
            await self._sample_edge()
            data_rx.shift(self.spi.miso)
        await self._shift_edge()
        return data_rx.data()

    async def transaction(
        self, send_data: BitVector, receive_len: int = 0, receive_offset=0, cs=None
    ):
        total_len = len(send_data) + receive_len + receive_offset
        assert not is_qualified(
            total_len
        ), "receive_len and receive_offset must be runtime constant"
        assert len(send_data) >= -receive_offset
        assert len(send_data) <= total_len

        shift_width = (
            None if instance_check(self.spi.mosi, Bit) else self.spi.mosi.width
        )

        self._start_transaction(cs)

        cnt = Signal[Unsigned.upto(total_len)](total_len)
        shift_out = OutShiftRegister(send_data, msb_first=True, unchecked=True)

        if receive_len != 0:
            shift_in = InShiftRegister(receive_len, msb_first=True, unchecked=True)

        self.spi.mosi <<= shift_out.shift(shift_width)

        while cnt:
            cnt <<= cnt - 1
            await self._sample_edge()

            if receive_len != 0:
                shift_in.shift(self.spi.miso)

            await self._shift_edge()
            self.spi.mosi <<= shift_out.shift(shift_width)

        self._end_transaction()

        if receive_len != 0:
            return shift_in.data()
        return None

    async def parallel_transaction(
        self, send_data: BitVector, shift_cnt: int | Unsigned, cs=None
    ):
        shift_width = (
            None if instance_check(self.spi.mosi, Bit) else self.spi.mosi.width
        )

        self._start_transaction(cs)

        cnt = Signal[Unsigned.upto(max_int(shift_cnt))](shift_cnt)
        shift_out = OutShiftRegister(send_data, msb_first=True, unchecked=True)
        shift_in = InShiftRegister(send_data.width, msb_first=True, unchecked=True)

        self.spi.mosi <<= shift_out.shift(shift_width)

        while cnt:
            cnt <<= cnt - 1
            await self._sample_edge()

            shift_in.shift(self.spi.miso)

            await self._shift_edge()
            self.spi.mosi <<= shift_out.shift(shift_width)

        self._end_transaction()

        return shift_in.data()

    async def transaction_context(
        self, send_data=None, cs: Bit | BitVector | None = None
    ) -> _SpiTransaction:
        self._start_transaction(cs)

        if send_data is not None:
            await self._send_data(as_bitvector(send_data))
        return _SpiTransaction(self)



#
#
#

from .reg import *
from .reg import _primitive_as_int
from cohdl.utility.code_writer import TextBlock


class Content:

    def __init__(self):
        self.name = None
        self.desc = None
        self.sw = None
        self.hw = None
        self.encode = None

        self.sub_content = []

    def to_list(self):
        result = []

        if self.name is not None:
            result.append(f'name = "{self.name}";')

        if self.desc is not None:
            result.append(f'desc = "{self.desc}";')

        if self.sw is not None:
            result.append(f"sw = {self.sw};")

        if self.hw is not None:
            result.append(f"hw = {self.hw};")

        if self.encode is not None:
            result.append(f"encode = {self.encode};")

        result.extend(self.sub_content)

        return result

    def append(self, raw):
        self.sub_content.append(raw)

    def add_description(self, input: type, metadata: list | None = None):
        if metadata is not None:
            for entry in metadata:
                if isinstance(entry, str):
                    self.desc = entry.strip()
                    return

        for elem in input.mro():
            if elem is object:
                return
            if elem.__doc__ is not None:
                self.desc = elem.__doc__.strip()
                return

    def add_meta(self, metadata: type | None):
        if metadata is None:
            return

        for info in metadata:
            if isinstance(info, str):
                self.desc = info
            elif isinstance(info, Access):
                self.sw = info.name
            elif isinstance(info, HwAccess):
                self.hw = info.name


class ComponentBlock(TextBlock):
    def __init__(
        self,
        typename: str,
        component_name: str | None = None,
        content: Content = [],
        trailer="};",
    ):
        if component_name is None:
            title = f"{typename} {{"
        else:
            title = f"{typename} {component_name} {{"

        super().__init__(title=title, content=content.to_list(), trailer=trailer)


def _find_enums(root: RegFile, result_set: set):
    for subtype in root._member_types_.values():
        if issubclass(subtype, RegFile):
            _find_enums(subtype, result_set)
        elif issubclass(subtype, Register):
            for fieldtype in subtype._field_types_.values():
                assert issubclass(fieldtype, FieldBase)
                arg = fieldtype._field_arg

                if arg.is_enum():
                    result_set.add(arg.underlying)


def _add_description(input: type, content: list, metadata: list | None = None):
    if metadata is not None:
        for entry in metadata:
            if isinstance(entry, str):
                content.append(f'desc = "{entry.strip()}";')
                return

    for elem in input.mro():
        if elem is object:
            return
        if elem.__doc__ is not None:
            content.append(f'desc = "{elem.__doc__.strip()}";')
            return


def _add_meta_args(metadata: type, content: list):
    if metadata is None:
        return

    desc = None
    sw_access = None
    hw_access = None

    for info in metadata:
        if isinstance(info, str):
            desc = info
        elif isinstance(info, Access):
            sw_access = info.name
        elif isinstance(info, HwAccess):
            hw_access = info.name

    if desc is not None:
        content.append(f'desc = "{desc}";')

    if sw_access is not None:
        content.append(f"sw = {sw_access};")

    if hw_access is not None:
        content.append(f"hw = {hw_access};")


def _impl_to_system_rdl(input, name: str | None = None, metadata=None):
    escape = _systemrdl_escape

    if issubclass(input, RegFile):
        content = Content()

        member_metadata = input._metadata_

        if name is None:
            content.name = escape(input.__name__.split("[")[0])
        else:
            content.name = escape(name)

        content.add_description(input, metadata)

        if issubclass(input, AddrMap):
            content.append(f"default regwidth = {input._register_tools_._word_width_};")
            content.append("")

            enumerations = set()
            _find_enums(input, enumerations)

            for enumtype in enumerations:
                content.append(_impl_to_system_rdl(enumtype))
            trailer = "};"
        else:
            trailer = "} " + f"{escape(name)} @ 0x{input._parent_offset_:0x};"

        for member_name, member_type in input._member_types_.items():
            content.append("")
            content.append(
                _impl_to_system_rdl(
                    member_type, member_name, member_metadata.get(member_name, None)
                )
            )
            content.append("")

        return ComponentBlock("addrmap", content.name, content=content, trailer=trailer)
    elif issubclass(input, Register):
        content = Content()
        content.add_description(input, metadata)

        for field_name, field_type in input._field_types_.items():
            content.append("")
            content.append(
                _impl_to_system_rdl(
                    field_type, field_name, input._metadata_.get(field_name, None)
                )
            )

        return ComponentBlock(
            "reg",
            content=content,
            trailer="}" + f" {escape(name)} @ 0x{input._parent_offset_:0x};",
        )
    elif issubclass(input, Field):
        content = Content()
        content.add_meta(metadata)

        arg = input._field_arg

        trailer_loc = f"[{arg.offset+arg.width-1}:{arg.offset}]"

        if arg.is_enum():
            content.encode = escape(arg.underlying.__name__)

        if arg.default is None:
            trailer_default = ""
        else:
            underlying = input._cohdlstd_underlying()
            default = underlying(arg.default)

            if issubclass(underlying, StdEnum):
                default = _primitive_as_int(default.raw)
            else:
                default = _primitive_as_int(default)

            trailer_default = f" = {escape(default)}"

        return ComponentBlock(
            "field",
            content=content,
            trailer="}" + f" {escape(name)} {trailer_loc}{trailer_default};",
        )
    elif issubclass(input, FlagField):
        content = Content()
        content.add_meta(metadata)

        arg = input._field_arg
        trailer_loc = f"[{arg.offset}:{arg.offset}]"

        return ComponentBlock(
            "field",
            content=content,
            trailer="}" + f" {escape(name)} {trailer_loc} = 0;",
        )
    elif issubclass(input, StdEnum):
        content = Content()
        enum_name = input.__name__.split("[")[0]

        for enumerator_name, enumerator in input.__members__.items():
            enum_def = f"{_systemrdl_escape(enumerator_name)} = {_primitive_as_int(enumerator.raw)}"

            info = enumerator.info

            if info is None:
                enum_def += ";"
            else:
                enum_def += f' {{ desc = "{info}"; }};'

            content.append(enum_def)

        return ComponentBlock("enum", _systemrdl_escape(enum_name), content=content)
    elif issubclass(input, (Input, Output, GenericRegister)):
        content = Content()

        content.add_description(input, metadata)
        content.add_meta(metadata)

        if issubclass(input, Input):
            access = f"sw = r;"
        else:
            access = f"sw = w;"

        content.append(f"field {{ {access} }} \\all [31:0];")

        return ComponentBlock(
            "reg",
            content=content,
            trailer="}" + f" {escape(name)} @ 0x{input._parent_offset_:0x};",
        )
    elif issubclass(input, Memory):
        content = Content()
        content.add_description(input, metadata)
        content.add_meta(metadata)

        if not input._writable_:
            content.sw = "r"

        content.append(f"mementries = {input._word_count_};")
        content.append(f"memwidth = {input._word_width_()};")

        return ComponentBlock(
            "external mem",
            content=content,
            trailer="}" + f" {escape(name)} @ 0x{input._parent_offset_:0x};",
        )

    raise AssertionError(f"invalid input {input}")


def to_system_rdl(input: AddrMap | type[AddrMap]):
    if not isinstance(input, type):
        input = type(input)

    return _impl_to_system_rdl(input).dump()


def _systemrdl_escape(inp: str):
    if inp in systemrdl_reserved:
        return f"\{inp}"
    return inp


systemrdl_reserved = {
    "abstract",
    "accesstype",
    "addressingtype",
    "addrmap",
    "alias",
    "all",
    "bit",
    "boolean",
    "bothedge",
    "compact",
    "component",
    "componentwidth",
    "constraint",
    "default",
    "encode",
    "enum",
    "external",
    "false",
    "field",
    "fullalign",
    "hw",
    "inside",
    "internal",
    "level",
    "longint",
    "mem",
    "na",
    "negedge",
    "nonsticky",
    "number",
    "onreadtype",
    "onwritetype",
    "posedge",
    "property",
    "r",
    "rclr",
    "ref",
    "reg",
    "regalign",
    "regfile",
    "rset",
    "ruser",
    "rw",
    "rw1",
    "signal",
    "string",
    "struct",
    "sw",
    "this",
    "true",
    "type",
    "unsigned",
    "w",
    "w1",
    "wclr",
    "woclr",
    "woset",
    "wot",
    "wr",
    "wset",
    "wuser",
    "wzc",
    "wzs",
    "wzt",
}



from __future__ import annotations


from ._intrinsic import _intrinsic
from ._primitive_type import _PrimitiveType


class _BooleanLiteral:
    _init_cnt = 0

    def __init__(self, value: bool):
        assert (
            _BooleanLiteral._init_cnt < 2
        ), "only two boolean literals (true and false) can exist"
        _BooleanLiteral._init_cnt += 1
        self._value = value

    def assign(self, other) -> None:
        raise AssertionError("assignment to literal type _BooleanLiteral impossible")

    @_intrinsic
    def __bool__(self) -> bool:
        return self._value

    def __await__(self):
        def coro():
            while not self._value:
                yield
            return self

        return coro()

    @_intrinsic
    def __call__(self, arg) -> _Boolean:
        if self is true:
            return _Boolean(arg)
        else:
            return _Boolean(not arg)


class MetaBoolean(type):
    def __str__(cls):
        return "Boolean"

    def __repr__(cls):
        return "Boolean"


class _Boolean(_PrimitiveType, metaclass=MetaBoolean):
    def __init__(self, value=False):
        self._value = bool(value)

    @property
    def type(self):
        return _Boolean

    def _assign(self, other) -> None:
        if isinstance(other, str):
            assert other in ("0", "1")
            self._value = other == "1"
        else:
            if isinstance(other, int):
                assert other in (0, 1)

            self._value = bool(other)

    def copy(self) -> _Boolean:
        return _Boolean(self._value)

    @_intrinsic
    def __bool__(self) -> bool:
        return self._value

    def __await__(self):
        def coro():
            while not self._value:
                yield
            return self

        return coro()


true = _BooleanLiteral(True)
false = _BooleanLiteral(False)
# TODO
boolean = _Boolean


# Null and Full are not synthesizable in the same sense as
# Bit or Bool. They only serve as helpers for initialization and assignment
# of other types.
# It is not possible to create Signals of type _NullType.


class _NullFullType:
    _init_cnt = 0

    def __init__(self):
        assert (
            _NullFullType._init_cnt <= 1
        ), "only two instances of _NullFullType (Null and Full) can exist"
        _NullFullType._init_cnt += 1

    @_intrinsic
    def __bool__(self):
        return self is Full

    @_intrinsic
    def __inv__(self):
        return Full if self is Null else Null

    @_intrinsic
    def __invert__(self):
        return Full if self is Null else Null


Null = _NullFullType()
"""
magic value that can be assigned to Bit/BitVector types
to set all bits to `0`
"""

Full = _NullFullType()
"""
magic value that can be assigned to Bit/BitVector types
to set all bits to `1`
"""



from __future__ import annotations
from ._intrinsic import _intrinsic
from ._primitive_type import _PrimitiveType


class Integer(_PrimitiveType):
    @staticmethod
    def decay(value: int | Integer) -> int:
        import cohdl

        if isinstance(value, int):
            assert not isinstance(value, bool)
            pass
        elif isinstance(value, Integer):
            value = value.get_value()
            assert isinstance(value, int)
        elif isinstance(value, (cohdl.Unsigned, cohdl.Signed)):
            value = value.to_int()

        assert isinstance(value, int), f"value is {value}, expected integer"

        return value

    def __init__(
        self,
        val: int | Integer | None = None,
        *,
        min: int | None = None,
        max: int | None = None,
    ):
        import cohdl

        self._min = min
        self._max = max

        if val is cohdl.Null:
            val = 0

        if val is None:
            self._val = 0
        else:
            assert isinstance(val, (int, Integer))
            self._val = int(val) if isinstance(val, int) else val.get_value()

    @property
    def type(self):
        return Integer

    def get_value(self) -> int:
        return self._val

    def copy(self):
        return Integer(self._val, min=self._min, max=self._max)

    def _assign(self, other: int | Integer):
        import cohdl

        if isinstance(other, Integer):
            other = other.get_value()
        elif isinstance(other, (cohdl.Unsigned, cohdl.Signed)):
            other = other.to_int()

        if self._min is not None:
            assert (
                self._min <= other
            ), f"assigned value {other} is smaller than the set minimum {self._min}"
        if self._max is not None:
            assert (
                self._max >= other
            ), f"assigned value {other} is larger than the set maximum {self._max}"

        assert isinstance(other, int), f"expected an integer argument not {other}"
        self._val = other

    def __hash__(self) -> int:
        return self._val

    @_intrinsic
    def __bool__(self) -> bool:
        return self._val != 0

    @_intrinsic
    def __index__(self) -> int:
        assert isinstance(
            self._val, int
        ), f"index expects an integer value not {self._val}"
        return self._val

    @_intrinsic
    def __add__(self, other: int | Integer) -> Integer:
        return Integer(self._val + Integer.decay(other))

    @_intrinsic
    def __radd__(self, other: int) -> Integer:
        return Integer(other + self._val)

    @_intrinsic
    def __sub__(self, other: int | Integer) -> Integer:
        return Integer(self._val - Integer.decay(other))

    @_intrinsic
    def __rsub__(self, other: int) -> Integer:
        return Integer(other - self._val)

    @_intrinsic
    def __and__(self, other: int | Integer) -> Integer:
        return Integer(self._val & Integer.decay(other))

    @_intrinsic
    def __or__(self, other: int | Integer) -> Integer:
        return Integer(self._val | Integer.decay(other))

    @_intrinsic
    def __xor__(self, other: int | Integer) -> Integer:
        return Integer(self._val ^ Integer.decay(other))

    @_intrinsic
    def __rand__(self, other: int) -> Integer:
        return Integer(self._val & other)

    @_intrinsic
    def __ror__(self, other: int) -> Integer:
        return Integer(self._val | other)

    @_intrinsic
    def __rxor__(self, other: int) -> Integer:
        return Integer(self._val ^ other)

    @_intrinsic
    def __floordiv__(self, other: int | Integer) -> Integer:
        return Integer(self._val // Integer.decay(other))

    @_intrinsic
    def __rfloordiv__(self, other: int) -> Integer:
        return Integer(other // self._val)

    @_intrinsic
    def __neg__(self) -> Integer:
        return Integer(-self._val)

    @_intrinsic
    def __pos__(self) -> Integer:
        return Integer(self._val)

    @_intrinsic
    def __eq__(self, other: int | Integer) -> bool:
        return bool(self._val == Integer.decay(other))

    @_intrinsic
    def __ne__(self, other) -> bool:
        return bool(self._val != Integer.decay(other))

    @_intrinsic
    def __gt__(self, other) -> bool:
        other = Integer.decay(other)
        assert self._val is not None, "integer value is not set"

        return bool(self._val > other)

    @_intrinsic
    def __lt__(self, other) -> bool:
        other = Integer.decay(other)
        assert self._val is not None, "integer value is not set"

        return bool(self._val < other)

    @_intrinsic
    def __ge__(self, other) -> bool:
        other = Integer.decay(other)
        assert self._val is not None, "integer value is not set"

        return bool(self._val >= other)

    @_intrinsic
    def __le__(self, other) -> bool:
        other = Integer.decay(other)
        assert self._val is not None, "integer value is not set"

        return bool(self._val <= other)

    @_intrinsic
    def __str__(self):
        return str(self._val)

    @_intrinsic
    def __repr__(self):
        return f"Integer({self._val}, min={self._min}, max={self._max})"



from __future__ import annotations

from ._bit_vector import BitVector, BitOrder
from ._integer import Integer

from ._intrinsic import _intrinsic
from cohdl.utility import Span


class Signed(BitVector):
    _SubTypes = {}

    @staticmethod
    @_intrinsic
    def _int_to_binary(width: int, number: int) -> str:
        return "".join([str(int(bool(number & 2**i))) for i in range(width)])[::-1]

    @staticmethod
    @_intrinsic
    def from_int(
        default: int | Integer,
        max: int | None = None,
        min: int | None = None,
        order: BitOrder = BitOrder.DOWNTO,
    ):
        if isinstance(default, Integer):
            default = default.get_value()

        if min is None:
            min = -abs(default)

        if max is None:
            max = abs(default)

        assert (
            min <= default <= max
        ), f"default value ({default}) outside representable range ({min}-{max})"

        width_neg = len(bin(abs(min) - 1)) - 2 - (min == 0)
        width_pos = len(bin(max)) - 2

        width = width_pos if width_neg < width_pos else width_neg

        return Signed[width + 1](default)

    @_intrinsic
    def __init__(
        self,
        val: None | BitVector | str | int = None,
    ):
        if isinstance(val, Integer):
            val = int(val)

        if isinstance(val, int):
            min = -(2 ** (self.width - 1))
            max = 2 ** (self.width - 1)
            assert (
                min <= val < max
            ), f"value {val} outside representable range ({min}-{max-1})"
            val = Signed._int_to_binary(self.width, val)

        super().__init__(val)

    @_intrinsic
    def pow_2(self, exp: int) -> Signed:
        if exp == 0:
            return Signed(self.width(), self, self._order)

        if exp > 0:
            return (self.as_bitvector() @ BitVector.zeros(exp)).as_signed()

        if exp >= self.width():
            return Signed(1, 0)

        return self.msb(self.width - exp).as_signed()

    @classmethod
    @_intrinsic
    def min_int(cls) -> int:
        return -(2 ** (cls.width - 1))

    @classmethod
    @_intrinsic
    def max_int(cls) -> int:
        return 2 ** (cls.width - 1) - 1

    @classmethod
    @_intrinsic
    def min(cls) -> Signed:
        return cls(cls.min_int())

    @classmethod
    @_intrinsic
    def max(cls) -> Signed:
        return cls(cls.max_int())

    @_intrinsic
    def _assign(self, other) -> None:
        import cohdl

        if isinstance(other, Integer):
            other = int(other)

        if isinstance(other, int):
            assert (
                self.min() <= other <= self.max()
            ), f"assigned value ({other}) outside representable range ({self.min()}-{self.max()})"

            if other < 0:
                binary = Span([*bin(abs(other + 1))[2:][::-1]]).generate(
                    lambda char: "0" if char == "1" else "1"
                )
                self._value.apply_zip(
                    lambda bit, char: bit._assign(char), binary.iter_extend("1")
                )
            else:
                binary = Span([*bin(other)[2:][::-1]])
                self._value.apply_zip(
                    lambda bit, char: bit._assign(char), binary.iter_extend("0")
                )
        elif isinstance(other, Signed):
            assert (
                self.width >= other.width
            ), "assigned value is wider than the target type"
            self._assign(other.to_int())
        elif isinstance(other, cohdl.Unsigned):
            assert (
                self.width > other.width
            ), "assigned unsigned value has equal or larger width than the signed target type"
            self._assign(other.to_int())
        elif (
            isinstance(other, (BitVector, str))
            or other is cohdl.Null
            or other is cohdl.Full
        ):
            super()._assign(other)
        else:
            raise AssertionError("invalid source type for assignment to signed value")

    @_intrinsic
    def to_int(self) -> int:
        result = 0

        for shift, bit in enumerate(self._value):
            if bit:
                result |= 1 << shift

        if self.msb():
            result -= 2**self.width

        return result

    @_intrinsic
    def add(self, rhs: Signed | int | Integer, target_width=None) -> Signed:
        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, int):
            rhs = Signed.from_int(rhs, order=self._order)

            if target_width is None:
                # adding integer, that cannot be represented
                # without explicit specification of target_width
                # is probably an error
                assert (
                    rhs.width() <= self.width()
                ), "added integer is not in the representable target range"
                target_width = self.width()
        else:
            assert self.order == rhs.order, "bitorder missmatch"
            assert isinstance(rhs, Signed), "expected signed argument"

            if target_width is None:
                target_width = max(self.width, rhs.width)

        target = Signed[target_width](0)

        carry = False

        sign_a = self._value[-1]
        sign_b = rhs._value[-1]

        for a, b, t in zip(
            self._value.iter_extend(sign_a),
            rhs._value.iter_extend(sign_b),
            target._value,
        ):
            a = bool(a)
            b = bool(b)

            t._assign((a + b + carry) & 1)
            carry = (a + b + carry) > 1

        return target

    @_intrinsic
    def sub(self, rhs: Signed | int | Integer, target_width=None) -> Signed:
        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        rhs = -rhs
        return self.add(rhs, target_width)

    @_intrinsic
    def __add__(self, rhs: Signed | int | Integer) -> Signed:
        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, int):
            rhs = Signed.from_int(rhs)

        if not isinstance(rhs, Signed):
            return NotImplemented

        return self.add(rhs)

    @_intrinsic
    def __radd__(self, lhs: int | Integer) -> Signed:
        if isinstance(lhs, Integer):
            lhs = lhs.get_value()

        return self.add(lhs)

    @_intrinsic
    def __sub__(self, rhs: Signed | int | Integer) -> Signed:
        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, int):
            rhs = Signed.from_int(rhs)

        if not isinstance(rhs, Signed):
            return NotImplemented

        return self.sub(rhs)

    @_intrinsic
    def __rsub__(self, lhs: int | Integer) -> Signed:
        if isinstance(lhs, Integer):
            lhs = lhs.get_value()

        return lhs + -self

    @_intrinsic
    def __mul__(self, rhs: Signed | int | Integer) -> Signed:
        result_width = self.width

        if isinstance(rhs, Signed):
            result_width = self.width + rhs.width
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        return Signed[result_width](lhs * rhs)

    @_intrinsic
    def __floordiv__(self, rhs: Signed | int | Integer) -> Signed:
        result_width = self.width

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        if rhs == 0:
            return Signed[result_width]()
        return Signed[result_width](lhs // rhs)

    @_intrinsic
    def __mod__(self, rhs: Signed | int | Integer) -> Signed:
        result_width = self.width

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        if rhs == 0:
            return Signed[result_width]()
        return Signed[result_width](lhs % rhs)

    @_intrinsic
    def __lshift__(self, rhs) -> Signed:
        width = self.width
        val = (self.to_int() << int(rhs)) & ((1 << width) - 1)

        if val & (1 << (width - 1)):
            val = val - (1 << (width))

        return Signed[self.width](val)

    @_intrinsic
    def __rshift__(self, rhs) -> Signed:
        val = self.to_int() >> int(rhs)
        return Signed[self.width](val)

    @_intrinsic
    def __abs__(self) -> Signed:
        if self >= 0:
            return Signed[self.width](self.to_int())
        else:
            return -self

    @_intrinsic
    def __neg__(self) -> Signed:
        return ~self + 1

    @_intrinsic
    def __eq__(self, other: BitVector | int | Integer) -> bool:
        if isinstance(other, Integer):
            other = other.get_value()

        if isinstance(other, int):
            return bool(other == self.to_int())
        return super().__eq__(other)

    @_intrinsic
    def __ne__(self, other: BitVector | int | Integer) -> bool:
        if isinstance(other, Integer):
            other = other.get_value()

        if isinstance(other, int):
            return bool(other != self.to_int())
        return super().__ne__(other)

    @_intrinsic
    def __lt__(self, rhs: Signed | int | Integer) -> bool:
        assert isinstance(rhs, (Signed, int, Integer)), f"invalid argument '{rhs}'"

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()

        return bool(self.to_int() < rhs)

    @_intrinsic
    def __gt__(self, rhs: Signed | int | Integer) -> bool:
        assert isinstance(rhs, (Signed, int, Integer)), f"invalid argument '{rhs}'"

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()

        return bool(self.to_int() > rhs)

    @_intrinsic
    def __le__(self, rhs: Signed | int | Integer) -> bool:
        assert isinstance(rhs, (Signed, int, Integer)), f"invalid argument '{rhs}'"

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()

        return bool(self.to_int() <= rhs)

    @_intrinsic
    def __ge__(self, rhs: Signed | int | Integer) -> bool:
        assert isinstance(rhs, (Signed, int, Integer)), f"invalid argument '{rhs}'"

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, Signed):
            rhs = rhs.to_int()

        return bool(self.to_int() >= rhs)

    @_intrinsic
    def resize(self, target_width: int | None = None, *, zeros: int = 0):
        if target_width is None:
            target_width = self.width + zeros

        assert (
            self.width + zeros <= target_width
        ), f"width of zero extended value ({self.width}+{zeros}) exceeds target width ({target_width})"
        val = self.to_int() * 2**zeros
        return Signed[target_width](val)

    @_intrinsic
    def __repr__(self):
        return f"Signed[{self.width-1}:0]({self._bit_str()})"

    @_intrinsic
    def __str__(self):
        return f"{self.to_int()}s"



from __future__ import annotations

import enum


import typing
from typing import cast

from cohdl.utility import Span

from cohdl._core._bit import Bit, BitState
from cohdl._core._integer import Integer
from cohdl._core._boolean import _NullFullType, Null
from cohdl._core._intrinsic import _intrinsic
from cohdl._core._primitive_type import _PrimitiveType

Self = typing.TypeVar("Self")


class BitOrder(enum.Enum):
    DOWNTO = enum.auto()
    UPTO = enum.auto()


class _BitVector(type):
    _width: int
    _order: BitOrder
    _SubTypes: dict[tuple, type]

    @_intrinsic
    def __getitem__(cls, size: int | slice):
        if isinstance(size, slice):
            assert size.step is None, "step parameter not allowed in slice argument"
            assert isinstance(size.start, int), "start parameter must be integer"
            assert isinstance(size.stop, int), "stop parameter must be integer"

            if size.start == 0:
                assert size.stop >= 0, "stop value cannot be negative"
                order = BitOrder.UPTO
                width = size.stop + 1
            elif size.stop == 0:
                assert size.start >= 0, "start value cannot be negative"
                order = BitOrder.DOWNTO
                width = size.start + 1
            else:
                raise AssertionError(
                    "invalid BitVector declaration, start or stop must be 0"
                )
        else:
            assert isinstance(size, int)
            assert size >= 0, "vector width must be positive"

            order = BitOrder.DOWNTO
            width = size

        shape = (order, width)

        if shape in cls._SubTypes:
            return cls._SubTypes[shape]

        if cls._SubTypes is BitVector._SubTypes:
            new_type = type(cls.__name__, (cls,), {"_width": width, "_order": order})
        else:
            new_type = type(
                cls.__name__,
                (cls, BitVector[width]),
                {"_width": width, "_order": order},
            )

        cls._SubTypes[shape] = new_type
        return new_type

    @_intrinsic
    def __str__(cls):
        if not hasattr(cls, "_width"):
            return f"{cls.__name__}"

        order = getattr(cls, "_order")
        width = getattr(cls, "_width")

        if order is BitOrder.DOWNTO:
            if width == 0:
                start = 0
                stop = ""
            else:
                start = width - 1
                stop = 0
        else:
            if width == 0:
                start = ""
                stop = 0
            else:
                start = 0
                stop = width - 1

        return f"{cls.__name__}[{start}:{stop}]"

    @classmethod
    def __repr__(cls):
        return cls.__str__(cls)

    @property
    def width(cls):
        return cls._width

    @property
    def order(cls):
        return cls._order


class BitVector(_PrimitiveType, metaclass=_BitVector):
    _SubTypes = {}
    _width: int
    _order: BitOrder

    @property
    def width(self):
        return self._width

    @property
    def order(self):
        return self._order

    @_intrinsic
    def __init__(
        self,
        val: None | BitVector | str | _NullFullType | Span[Bit] = None,
    ):
        if isinstance(val, Span):
            # init bitvector as a span over a subrange of an existing bitvector
            assert self._width == len(val)
            self._value = val
        else:
            # init bitvector as a new parent object

            self._value: Span[Bit] = Span([Bit() for _ in range(self._width)])

            if val is None:
                # no init of _value required
                return

            if isinstance(val, BitVector):
                assert (
                    val._width == self._width
                ), f"width mismatch in vector constructor source-width({val._width}) != target-width({self._width})"
                start_val = val._value
            elif isinstance(val, str):
                assert len(val) == self._width
                start_val = Span.from_iter(val[::-1], Bit)
            elif isinstance(val, _NullFullType):
                start_val = [Bit(val) for _ in range(self._width)]
            else:
                raise AssertionError("invalid default value")

            self._value.apply_zip(
                lambda bit, state: bit._assign(state),
                cast(Span[Bit], start_val),
            )

    @_intrinsic
    def _is_uninitialized(self):
        return all(bit._is_uninitialized() for bit in self._value)

    @_intrinsic
    def __iter__(self):
        return iter(self._value)

    @_intrinsic
    def _bit_str(self) -> str:
        return "".join([str(bit) for bit in self._value][::-1])

    @_intrinsic
    def copy(self):
        return type(self)(Span.from_iter(self._value, lambda bit: bit.copy()))

    @_intrinsic
    def right(self, width=None, rest=None) -> BitVector | Bit:
        if width is None and rest is None:
            return self._value[0]

        if (width is not None) and (rest is not None):
            assert (
                width + rest == self._width
            ), f"width + rest does not equal the vectors width ({width}+{rest} != {self._width})"

        if rest is not None:
            width = self._width - rest
        elif width is None:
            width = 1

        width = int(width)

        assert (
            1 <= width <= self._width
        ), f"invalid subvector width {width} for vector of width {self._width}"

        return BitVector[width](self._value[0:width])

    @_intrinsic
    def left(self, width=None, rest=None) -> BitVector | Bit:
        if width is None and rest is None:
            return self._value[-1]

        if (width is not None) and (rest is not None):
            assert (
                width + rest == self._width
            ), f"width + rest does not equal the vectors width ({width}+{rest} != {self._width})"

        if rest is not None:
            width = self._width - rest
        elif width is None:
            width = 1

        width = int(width)

        assert (
            1 <= width <= self._width
        ), f"invalid subvector width {width} for vector of width {self._width}"

        return BitVector[width](self._value[self._width - width :])

    @_intrinsic
    def lsl(self, cnt=None, fill=None):
        if fill is not None:
            assert cnt is None or cnt is len(fill)
            fill_val = fill
        else:
            if cnt is None:
                fill_val = Bit(0)
            else:
                fill_val = BitVector[cnt](Null)

        return self.right(rest=len(fill_val)) @ fill_val

    @_intrinsic
    def lsr(self, cnt=None, fill=None):
        if fill is not None:
            assert cnt is None or cnt is len(fill)
            fill_val = fill
        else:
            if cnt is None:
                fill_val = Bit(0)
            else:
                fill_val = BitVector[cnt](Null)

        return fill_val @ self.left(rest=len(fill_val))

    @_intrinsic
    def lsb(self, width: int | None = None, rest=None) -> BitVector | Bit:
        if self._order is BitOrder.DOWNTO:
            return self.right(width, rest)
        return self.left(width, rest)

    @_intrinsic
    def msb(self, width=None, rest=None) -> BitVector | Bit:
        if self._order is BitOrder.DOWNTO:
            return self.left(width, rest)
        return self.right(width, rest)

    @_intrinsic
    def _assign(self, other):
        if isinstance(other, _NullFullType):
            other = type(self)(other)
        if isinstance(other, str):
            other = BitVector[len(other)](other)

        assert isinstance(other, BitVector), f"expected BitVector not '{other}'"

        assert (
            self._width == other._width
        ), f"width mismatch in vector assignment (target_width({self._width}) != source_width({other._width}))"

        self._value.apply_zip(lambda bit, other: bit._assign(other), other._value)

    @_intrinsic
    def __bool__(self) -> bool:
        return any([bit for bit in self._value])

    @_intrinsic
    def __invert__(self):
        return type(self)(Span.from_iter(self._value, lambda bit: ~bit))

    @_intrinsic
    def __inv__(self):
        return self.__invert__()

    @_intrinsic
    def __or__(self, other):
        if not isinstance(other, BitVector):
            return NotImplemented

        Self = type(self)
        Other = type(other)

        assert Self is Other, f"type mismatch in binary or ({Self} != {Other})"

        return Self(
            Span.from_zip_iter(self._value, other._value, modifyer=lambda a, b: a | b)
        )

    @_intrinsic
    def __and__(self, other):
        if not isinstance(other, BitVector):
            return NotImplemented

        Self = type(self)
        Other = type(other)

        assert Self is Other, f"type mismatch in binary and ({Self} != {Other})"

        return Self(
            Span.from_zip_iter(self._value, other._value, modifyer=lambda a, b: a & b)
        )

    @_intrinsic
    def __xor__(self, other):
        if not isinstance(other, BitVector):
            return NotImplemented

        Self = type(self)
        Other = type(other)

        assert Self is Other, f"type mismatch in binary xor ({Self} != {Other})"

        return Self(
            Span.from_zip_iter(self._value, other._value, modifyer=lambda a, b: a ^ b)
        )

    @_intrinsic
    def __eq__(self, other: BitVector) -> bool:
        if isinstance(other, _NullFullType):
            if other:
                return all(bit for bit in self._value)
            else:
                return not any(bit for bit in self._value)

        if isinstance(other, str):
            other = BitVector[len(other)](other)

        assert self.width == other.width, "cannot compare bitvectors of different width"

        for a, b in zip(self._value, other._value):
            if bool(a) != bool(b):
                return False

        return True

    @_intrinsic
    def __ne__(self, other: BitVector) -> bool:
        return not self.__eq__(other)

    @_intrinsic
    def __matmul__(self, rhs: Bit | BitVector) -> BitVector:
        if isinstance(rhs, Bit):
            result = BitVector[self._width + 1]()

            result._value[0]._assign(rhs)
            result._value[1:].apply_zip(
                lambda bit, other: bit._assign(other), self._value
            )

            return result
        elif isinstance(rhs, BitVector):
            result = BitVector[self.width + rhs.width]()

            result._value[rhs.width :].apply_zip(
                lambda bit, other: bit._assign(other), self._value
            )
            result._value[: rhs.width].apply_zip(
                lambda bit, other: bit._assign(other), rhs._value
            )

            return result
        else:
            return NotImplemented

    @_intrinsic
    def __rmatmul__(self, lhs: Bit) -> BitVector:
        assert isinstance(lhs, Bit)

        result = BitVector[self._width + 1]()
        result._value[self._width]._assign(lhs)
        result._value[0 : self._width].apply_zip(
            lambda bit, val: bit._assign(val), self._value
        )

        return result

    @_intrinsic
    def __len__(self) -> int:
        return len(self._value)

    @_intrinsic
    def subvector(self, offset: int, width: int | None = None):
        if width is None:
            width = self.width() - offset

        assert offset >= 0 and offset + width <= self.width()

        if width == 0:
            return BitVector[0]()

        return BitVector[width](self._value)

    @_intrinsic
    def __getitem__(self, key: int | Integer | slice) -> BitVector | Bit:
        if isinstance(key, (tuple, list)):
            first, *rest = key

            if len(rest) == 0:
                if isinstance(first, slice):
                    return self.__getitem__(first)
                else:
                    return self.__getitem__(slice(first, first))
            else:
                return self.__getitem__(first) @ self.__getitem__(rest)
        elif isinstance(key, (int, Integer)):
            key = Integer.decay(key)
            assert 0 <= key < self.width, "index exceeds vector width"
            return self._value[key]
        elif isinstance(key, slice):
            assert key.step is None, "step parameter cannot be used"
            start = Integer.decay(key.start)
            stop = Integer.decay(key.stop)

            width = max(start, stop) - min(start, stop) + 1

            if stop <= start:
                return BitVector[width](self._value[stop : start + 1])
            else:
                raise RuntimeError("not implemented")
        else:
            raise RuntimeError(f"invalid slice parameter <{key}>")

    @_intrinsic
    def __setitem__(self, key: int | Integer | slice, value: Bit | BitVector):
        if not isinstance(key, slice):
            assert isinstance(value, (Bit, BitState))
            self._value[int(key)].assign(value)
        else:
            assert key.step is None, "step parameter cannot be used"
            start = int(key.start)
            stop = int(key.stop)

            assert isinstance(value, BitVector)

            start, stop = sorted([start, stop])

            width = stop - start + 1
            assert len(value) == width

            self._value[start : stop + 1].apply_zip(
                lambda bit, other: bit.assign(other), value._value
            )

    def __hash__(self) -> int:
        return len(self) + 1000

    def __str__(self) -> str:
        return self._bit_str()

    def __repr__(self) -> str:
        return f"{type(self)}({self._bit_str()})"

    @_intrinsic
    @property
    def unsigned(self: BitVector):
        from . import Unsigned

        return Unsigned[self._width](self._value)

    @_intrinsic
    @unsigned.setter
    def unsigned(self: BitVector, value):
        self.unsigned._assign(value)

    @_intrinsic
    @property
    def signed(self):
        from . import Signed

        return Signed[self._width](self._value)

    @_intrinsic
    @signed.setter
    def signed(self: BitVector, value):
        self.signed._assign(value)

    @_intrinsic
    @property
    def bitvector(self) -> BitVector:
        return BitVector[self._width](self._value)

    @_intrinsic
    @bitvector.setter
    def bitvector(self: BitVector, value):
        self.bitvector._assign(value)



from __future__ import annotations

from ._intrinsic import _intrinsic


class _PrimitiveType:
    """
    all types, that can be wrapped in a type qualifier
    derive from this type
    """

    @_intrinsic
    def __ilshift__(self, other):
        raise AssertionError(
            f"primitive type '{type(self)}' must be type qualified to support assignment"
        )

    @_intrinsic
    def __imatmul__(self, other):
        raise AssertionError(
            f"primitive type '{type(self)}' must be type qualified to support assignment"
        )

    @_intrinsic
    def __ixor__(self, other):
        raise AssertionError(
            f"primitive type '{type(self)}' must be type qualified to support assignment"
        )

    @property
    def next(self):
        raise AssertionError(
            "attempt to use 'next' property on a primitive type, a type qualifier is required"
        )

    @property
    def value(self):
        raise AssertionError(
            "attempt to use 'value' property on a primitive type, a type qualifier is required"
        )

    @property
    def push(self):
        raise AssertionError(
            "attempt to use 'push' property on a primitive type, a type qualifier is required"
        )

    @next.setter
    def next(self, value):
        raise AssertionError(
            "attempt to use 'next' property on a primitive type, a type qualifier is required"
        )

    @next.setter
    def value(self, value):
        raise AssertionError(
            "attempt to use 'value' property on a primitive type, a type qualifier is required"
        )

    @next.setter
    def push(self, value):
        raise AssertionError(
            "attempt to use 'push' property on a primitive type, a type qualifier is required"
        )


def is_primitive_type(t):
    if not isinstance(t, type):
        return False
    else:
        return issubclass(t, _PrimitiveType)


def is_primitive(obj):
    return isinstance(obj, _PrimitiveType)
